from .search_space_iterator import *

def range_op(rangez,default_value=0.,f_inner=np.subtract,f_outer=np.add):
    rangez_ = to_noncontiguous_ranges(rangez)
    l = len(rangez_)
    s = default_value
    for i in range(0,l):
        f_out = f_inner(rangez_[i][1],rangez_[i][0])
        s = f_outer(s,f_out)
    return s

"""
float accumulator operation over vector `v`
"""
def vec_op(v,interop,default_value=0.):
    s = default_value
    for v_ in v: 
        s = interop(s,v_)
    return s


############### a lightweight standard n-dimensional index incrementor
############### built on top of <SearchSpaceIterator>. 
##################################################################

def not_permutation_condition(v):
    stat = True
    l = len(v) - 1
    for i in range(l):
        stat = v[i] <= v[i+1]
        if not stat: break 
    return stat 

def is_reflective_condition(v):
    return len(v) == len(np.unique(v))

"""
Generates numbers in an n-dimensional square
structure according to variables such as 
`is_perm`, `is_reflective`. The variable 
`subset_size` indicates the dimension of the 
output from `__next__`. 

The value `length` is the length of the vector
that the output of this class is supposed to 
accomodate. 
"""
class BatchIncrStruct:

    def __init__(self,length:int,is_perm=False,\
        is_reflective=False,subset_size=2):
        """
        length := int or float, specifies the max 
                for each bound
        subset_size := int, otherwise known as the n in n-dim.
        """

        assert type(length) == type(subset_size)
        assert type(length) == int
            
        assert type(is_reflective) == type(is_perm)
        assert type(is_reflective) == bool

        self.l = length
        self.is_perm = is_perm
        self.is_ref = is_reflective
        self.subset_size = subset_size
        self.at_i = np.zeros((subset_size,))
        self.preprocess()
        return

    def preprocess(self):
        bounds = np.zeros((self.subset_size,2))
        bounds[:,1] = self.l

        column_order = [i for i in range(self.subset_size)][::-1]
        ssi_hop = self.l
        cycle_on = False
        cycle_is = 0
        self.ssi = SearchSpaceIterator(bounds,self.at_i,\
            column_order,ssi_hop,cycle_on,cycle_is)

    """
    end value is `None` instance. 
    """
    def __next__(self):
        stat = False
        while not stat: 
            q,stat = self.next_() 
        return q 

    def next_(self): 
        if self.ssi.reached_end():
            return None,True 

        v = next(self.ssi)
        v,stat = self.pass_next(v) 
        if type(v) != type(None): 
            v = np.asarray(v,dtype=int)

        return v,stat 

    def pass_next(self,v):

        stat = True if type(v) != type(None) else False
        if not stat: return v,stat 

        if not is_reflective_condition(v) and not self.is_ref:
            stat = False

        if not not_permutation_condition(v) and not self.is_perm:
            stat = False

        return v,stat

#################### methods for calculating coverage of sequence
###################################################

"""
maps a sequence of ranges to one, with 
noncontiguous ranges and has the greatest
similarity to original `rangez`. 

"""
def to_noncontiguous_ranges(rangez,is_sorted=False):
    if not is_sorted:
        rangez = sorted(rangez,key=lambda x: x[0])
    
    i = 0
    while i < len(rangez) - 1:

        r1 = rangez[i]
        r2 = rangez[i+1]
        # check if intersection
        stat = r1[1] >= r2[0] and r1[1] <= r2[1]
        # case: intersection
        if stat:
            r3 = [r1[0],r2[1]]
            rangez.pop(i)
            rangez.pop(i)
            rangez.insert(i,r3)
        else: 
            i += 1
    return rangez

def floatseq_to_rangeseq(vf,rv,max_radius:float):
    assert len(rv) == 2 and rv[0] <= rv[1]
    assert max_radius > 0
    r = []
    for f in vf:
        min_f = max(rv[0],f-max_radius)
        max_f = min(rv[1],f+max_radius)
        r.append([min_f,max_f])
    return np.array(r) 

"""
measure does not take into account improper bounds (see 
the section of library on functions `*improper_bounds*`)
"""
def coverage_of_sequence(vf,rv,max_radius:float):
    rs = floatseq_to_rangeseq(vf,rv,max_radius)
    q = range_op(rs,default_value=0.,f_inner=np.subtract,f_outer=np.add)
    qmax = rv[1] - rv[0]
    return np.round(q/qmax,5)

############ methods for calculating unidirectional point distance
#######################################################

def normalized_float_uwpd(v,rv):
    l = len(v)
    d = uwpd(v,accum_op=lambda x1,x2: x1 + x2)
    dmax = max_float_uwpd(l,rv) 
    return np.round(d/dmax,5)

def uwpd(v,pairwise_op=lambda x1,x2: np.abs(x2 - x1),accum_op=None):
    assert type(v) == np.ndarray
    assert v.ndim == 1

    # using the 
    bis = BatchIncrStruct(len(v),False,False,2)
    stat = True 
    stat2 = type(accum_op) == type(None)
    result = [] if stat2 else 0.0
    while stat:
        index = next(bis)
        stat = not (type(index) == type(None))

        if not stat: 
            continue 

        op_val = pairwise_op(v[index[0]],v[index[1]])

        if stat2:
            result.append(op_val)
            continue
        result = accum_op(result,op_val) 
    return result

def max_float_uwpd(l,fr):
    assert len(fr) == 2 and fr[0] <= fr[1]
    s = 0
    for i in range(0,l): s += i
    return (fr[1] - fr[0]) * s  

###################################################

class APRNGGauge: 

    """
    aprng := Abstract Pseudo random generator function; values are called by 
             `aprng()`.
    frange := float range
    """
    def __init__(self,aprng,frange,pradius:float):
        assert len(frange) == 2
        assert frange[0] <= frange[1] 

        self.aprng = aprng
        self.frange = frange
        self.frange_default = frange 
        self.pradius = pradius
        # n * 2 array; 
        ## two columns
        ## [0]: cycle coverage
        ## [1]: unidirectional weighted point distance measure
        self.measurements = []
        self.cycle = None
        return 

    def assign_cycle(self,cycle):
        self.cycle = cycle

    def measure_cycle(self,max_size=100000,\
        term_func=lambda l,l2: type(l) == type(None),\
        auto_frange:bool=False):

        if type(self.cycle) == type(None):
            self.cycle = self.cycle_one(max_size,term_func)
        assert self.cycle.ndim == 1

        # get the coverage
        if auto_frange:
            m1,m2 = min(self.cycle),max(self.cycle)
            self.frange = [m1,m2] 

        cov = coverage_of_sequence(self.cycle,self.frange,self.pradius)

        # get the normed uwpd
        nuwpd = normalized_float_uwpd(self.cycle,self.frange)
        self.measurements.append([cov,nuwpd])
        return [cov,nuwpd]

    def cycle_one(self,max_size:int,term_func):
        start = self.aprng()
        c = [start]
        stat = True
        l,prev = None,None 

        while len(c) < max_size and stat:
            prev = l
            l = self.aprng()
            stat = term_func(l,prev)  

            # if terminate, 
            if stat: 
                c.append(l) 
        return np.array(c)

###################################################################################


"""
graph-to-tree decomposition
"""
from .graph_basics import * 
from .numerical_generator import prg_seqsort,prg_seqsort_ties,prg__constant,\
    prg__n_ary_alternator

class TNode: 

    def __init__(self,idn,next_cycling:bool=False,root_stat:bool=False,\
        rdistance:int = 0): 
        self.idn = idn 
        self.next_cycling = next_cycling
        self.children = []  
        self.cindex = 0 # used to traverse children node in search process 
        self.root_stat = root_stat 
        self.rdistance = rdistance 
        # children has been set? 
        self.children_set = False 
        self.xclist = [] 
        return

    """
    setattr is None|(varname,g); g a function called by g() 
    """
    @staticmethod
    def dfs(tn,display:bool=True,collect:bool=True,reset_index:bool=True,\
        set_attr=None,fetch_set=set()):

        def settr(tn_): 
            if type(set_attr) == type(None):
                return

            attrname = set_attr[0] 
            vx = set_attr[1]() 
            setattr(tn_,attrname,vx)
            return 

        if reset_index: tn.cindex = 0 
        cache = [tn] 
        d = defaultdict(set)
        mdepth = 0 
        fset = []  
        while len(cache) > 0: 
            t_ = cache.pop(0)
            settr(t_)

            if t_.idn in fetch_set:
                fset.append(t_)
            
            if display: 
                print(t_)
            cx = set([c.idn for c in t_.children])
            if collect: 
                d[t_.idn] = d[t_.idn] | cx 
            
            q = next(t_)
            if type(q) != type(None): 
                if reset_index: 
                    q.cindex = 0 
                cache.insert(0,t_)
                cache.insert(0,q) 
                mdepth = max([q.rdistance,mdepth])
            else: 
                if reset_index:
                    t_.cindex = 0 
        return d,mdepth,fset 

    # TODO: more thorough testing required. 
    @staticmethod
    def collate_keys(tn,is_bfs:bool=True,prg=None):
        tn.cindex = 0 
        q = [tn]
        K = [tn.idn]

        while len(q) > 0: 
            tn_ = q.pop(0) 
            if is_bfs: 
                q2 = [] 
                while True: 
                    tn2 = next(tn_)
                    if type(tn2) == type(None): break 
                    tn2.cindex = 0 
                    q2.append(tn2)
                if type(prg) != type(None): 
                    q2 = prg_seqsort(q2,prg)
                K.extend([q2_.idn for q2_ in q2])
                q.extend(q2) 
            else: 
                stat = True 
                if tn_.idn in K and type(prg) != type(None): 
                    stat = bool(prg() % 2)

                if stat: 
                    K.append(tn_.idn)

                tn2 = next(tn_)
                if type(tn2) == type(None): continue 
                tn2.cindex=0
                q.insert(0,tn_)
                q.insert(0,tn2)
        return K 

    @staticmethod 
    def size_count(tn): 
        q = [tn]
        c = 0 
        
        while len(q) > 0: 
            x = q.pop(0)
            x.cindex = 0 
            c += 1

            while True: 
                x2 = next(x) 
                if type(x2) == type(None): 
                    break 
                q.append(x2) 
        return c 


    def index_of_child(self,idn): 
        for (i,c) in enumerate(self.children): 
            if c.idn == idn: 
                return i 
        return -1 

    def add_children(self,cs): 
        for c in cs: 
            assert type(c) == TNode 
            self.children.append(c)
        self.children_set = True 

    def add_xclusion(self,xclude):
        assert type(xclude) == list 
        self.xclist.extend(xclude) 

    def __next__(self):
        if not self.next_cycling and self.cindex >= len(self.children): 
            return None 
        if len(self.children) == 0: return None 
        
        q = self.children[self.cindex % len(self.children)] 
        self.cindex += 1
        return q

    def __str__(self): 
        s = "idn:\t" + str(self.idn) + "\n"
        s += "rdistance:\t" + str(self.rdistance) + "  index:\t" \
            + str(self.cindex) + "\n"
        q = [str(c.idn) for c in self.children] 
        q = " ".join(q) 
        s += "children: " + q + "\n"
        s += "is root: " + str(self.root_stat) + "\n"
        return s 

class G2TDecomp: 

    def __init__(self,d,decomp_rootnodes=[],excl_mem_depth=float('inf'),\
            child_capacity=float('inf'),parent_capacity=float('inf'),prg=None): 
        """
        finds a directed acyclic graph (DAG) decomposition of a graph `d`. Every 
        directed acyclic graph is of a subclass of the category of DAG. For this 
        subclass of DAG, the siblings from a parent node cannot travel to each other.
        """

        assert type(d) == defaultdict
        assert d.default_factory == set 
        assert child_capacity > 0 and parent_capacity > 0

        graph_childkey_fillin(d)
        self.d = d 
        self.d_ = deepcopy(self.d)
        self.is_directed = is_directed_graph(d) 
        self.rn = decomp_rootnodes
        self.dr_map = defaultdict(int) 
        self.excl_mem_depth = excl_mem_depth
        self.cc = child_capacity
        self.pc = parent_capacity
        self.prg = prg  
        self.predecomp() 

        # vars used for dfs search 
        self.decompositions = [] 
        self.decomp_queue = [] 
            # stores skipped nodes for every dfs search 
            # by a root node 
        self.skipped_nodes = [] 
            # neighbor-parent degree map. used as reference to 
            # satisfy upper-threshold values set by `excl_mem_depth`,
            # `cc`, and `pc`. 
        self.cdeg_map = None 
        self.cdeg_map2 = None 

        self.fstat = False 

    def predecomp(self):
        if len(self.rn) > 0: 
            return 
        
        gd = GraphComponentDecomposition(self.d) 
        gd.decompose() 

        self.preset_depth_rank_map(gd)
        self.rn = [(k,v) for k,v in self.dr_map.items()] 
        if type(self.prg) == type(None): 
            self.rn = sorted(self.rn,key=lambda x:x[1]) 
        else: 
            vf = lambda x: x[1] 
            self.rn = prg_seqsort_ties(self.rn,self.prg,vf)
        self.rn = [k for (k,v) in self.rn]
        return

    def preset_depth_rank_map(self,gd): 
        if self.is_directed: 
            self.dr_map = gd.depth_rank_map() 
        else: 
            for k in self.d.keys(): 
                self.dr_map[k] = 0

    """
    main method 
    """
    def decompose(self): 
        while not self.fstat:
            next(self)

    def __next__(self): 
        if len(self.decomp_queue) == 0: 
            stat = self.next_tree() 
            if not stat: 
                self.fstat = True 
            return 

        self.next_node()
        
    def next_tree(self): 
        '''
        initializes a new tree, represented by class<TNode>. 
        '''
        if len(self.rn) == 0: 
            return False 

        x = self.rn.pop(0)
        tn = TNode(x,False,True,0)
        self.decompositions.append(tn) 
        self.decomp_queue.append(tn) 
        self.store_np_degrees() 
        return True 

    def set_children_for_node(self,tn):
        assert tn.children_set == False 

        # iterate through possible children 
        cx = children_of(self.d_,tn.idn)
        nx = doubly_connected(self.d_,tn.idn)
        cx = cx | nx 
        cx_ = [] 
        for q in cx: 
            if self.is_possible_child(q): 
                cx_.append(q)
        
        cx_ = set(cx_) - flatten_setseq(tn.xclist)

        # case: no children 
        if len(cx_) == 0: return False 

        # case: get the number of children based on `prg` or 
        #       default of `cc`. 
        cx_ = sorted(cx_) 
        cc = len(cx_) if float('inf') == self.cc else self.cc 
        if type(self.prg) != type(None): 
            cx_ = prg_seqsort(cx_,self.prg)
            cc = (self.prg() % cc) + 1 
        cx_ = cx_[:cc]
        self.init_children_for_node(tn,cx_)
        return True 

    def init_children_for_node(self,tn,cx_idn): 
        q = deepcopy(tn.xclist)

        # exclusion list for every child
        df = len(q) + 1 - self.excl_mem_depth
        if df > 0: 
            q = q[df:]
        q.append({tn.idn} | set(cx_idn))

        cxs = []
        E = [] 
        for cx in cx_idn: 
            q2 = deepcopy(q) 
            tn2 = TNode(cx,rdistance=tn.rdistance+1)
            tn2.add_xclusion(q2)
            cxs.append(tn2)
            E.append((tn.idn,cx))
            
        tn.add_children(cxs)
        delete_graph_edges(self.d_,E,is_directed=True)

    def is_possible_child(self,idn): 
        df = self.cdeg_map[idn] - self.cdeg_map2[idn]
        return df < self.pc 

    def next_node(self):
        tn = self.decomp_queue.pop(0) 
        if not tn.children_set: 
            self.set_children_for_node(tn)

        tn2 = next(tn) 
        if type(tn2) == type(None):
            tn.xclist.clear() 
            return False 

        self.decomp_queue.insert(0,tn)
        self.decomp_queue.insert(0,tn2)
        return

    #--------------------- conditional methods for next 

    def store_np_degrees(self): 
        q = np_degree_map(self.d_)
        self.cdeg_map = defaultdict(int) 
        for k,v in q.items(): 
            self.cdeg_map[k] += v[0] + v[1] 
        self.cdeg_map2 = deepcopy(self.cdeg_map)
        return


from morebs2 import search_space_iterator,matrix_methods
import numpy as np

def SearchSpaceIterator_case_1():

    bounds = np.array([[0,1],\
                    [0,1],\
                    [0,1]
        ])
    startPoint = np.array([0,1,0])

    # TODO: test w/ this column order
    columnOrder = [0,1,2]
    return search_space_iterator.SearchSpaceIterator(bounds, startPoint, columnOrder, 2)

def SearchSpaceIterator_case_2():
    bounds = np.array([[0,1],\
                    [0,1],\
                    [0,1]
        ])
    startPoint = np.array([0,0,0])

    # TODO: test w/ this column order
    columnOrder = [2,0,1]
    return search_space_iterator.SearchSpaceIterator(bounds, startPoint, columnOrder, 5)

def SearchSpaceIterator_case_3():
    bounds = np.array([[0,1],\
                    [0,1],\
                    [0,1]
        ])
    startPoint = np.array([0,0,0])

    # TODO: test w/ this column order
    columnOrder = [0,1,2]
    return search_space_iterator.SearchSpaceIterator(bounds, startPoint, columnOrder, 3)

def SearchSpaceIterator_case_4():
    bounds = np.array([[0,1],\
                    [0,1],\
                    [0,1],\
                    [0,1],\
                    [0,1]])

    startPoint = np.array([0.0,0.5,0.3, 1,0.75])
    columnOrder = [4,2,0,1,3]
    search_space_iterator.HopPattern.DEF_INCREMENT_RATIO = round(1/7,10)
    ##return SearchSpaceIterator(bounds, startPoint, columnOrder,\
    ##    7, "proportional")
    return search_space_iterator.SearchSpaceIterator(bounds, startPoint, columnOrder,7)

def SearchSpaceIterator_case_5():
    ssi = SearchSpaceIterator_case_4()
    hop_pattern.HopPattern.DEF_INCREMENT_RATIO = round(1/4,2)
    return ssi

def SearchSpaceIterator_case_6():

    bounds = np.array([[0,10],\
                    [0,10],\
                    [0,10.00],\
                    [0,10],\
                    [0,10.00]])

    startPoint = np.array([5.0, 4.0, 3.0, 8.0, 7.5])
    columnOrder = [4,2,0,1,3]
    ssih = 2
    return search_space_iterator.SearchSpaceIterator(bounds, startPoint, columnOrder, ssih,cycleIs = 1)

def SearchSpaceIterator_case_7():

    bounds = np.array([[0,10],\
                    [0,10],\
                    [0,10.00],\
                    [0,10],\
                    [0,10.00]])

    bounds = invert_bounds(bounds)
    startPoint = np.array([5.0, 4.0, 3.0, 8.0, 7.5])
    columnOrder = [4,2,0,1,3]
    ssih = 2
    return search_space_iterator.SearchSpaceIterator(bounds, startPoint, columnOrder, ssih)

"""
case is used to test out values given head in {0,1}
"""
def SearchSpaceIterator_case_8(head):

    bounds = np.array([[0,1],\
                    [0,1],\
                    [0,1],\
                    [0,1],\
                    [0,1]])

    startPoint = np.array([0,1.0,0.0,1.0,0.5])
    columnOrder = [4,3,2,1,0]
    ssih = 2
    return search_space_iterator.SearchSpaceIterator(bounds,startPoint,columnOrder,ssih, True,cycleIs = head)

def SearchSpaceIterator_case_9(): 

    bounds = np.array([[0,2],\
        [0,3],\
        [0,4],\
        [0,6],\
        [0,5],\
        [0,7]])

    start_point = np.array([0,0,0,0,0,0])
    column_order = [i for i in range(6)][::-1]
    ssihop = np.array([2,3,4,6,5,7])

    ssi = search_space_iterator.SearchSpaceIterator(bounds, start_point,\
        column_order, SSIHop = ssihop,cycleOn=False)
    return ssi 

def SkewedSearchSpaceIterator_args_1():

    twoThirds = round(2/3.0,5)
    oneThirds = round(1/3.0,5)

    bounds = np.array([[twoThirds,oneThirds],\
                    [twoThirds,oneThirds],\
                    [twoThirds,oneThirds],\
                    [twoThirds,oneThirds],\
                    [twoThirds,oneThirds]])

    parentBounds = np.array([[0,1.0],\
                    [0,1.0],\
                    [0,1.0],\
                    [0,1.0],\
                    [0,1.0]])

    return bounds,parentBounds

def SkewedSearchSpaceIterator_case_1():
    b,pb = SkewedSearchSpaceIterator_args_1()
    return search_space_iterator.SkewedSearchSpaceIterator(b,pb,None,None,SSIHop = 3, cycleOn = True, cycleIs = 1)

def SkewedSearchSpaceIterator_case_2():
    b,pb = SkewedSearchSpaceIterator_args_1()
    sp = b[:,1]
    return search_space_iterator.SkewedSearchSpaceIterator(b,pb,sp,None,SSIHop = 3, cycleOn = True, cycleIs = 1)

def SkewedSearchSpaceIterator_case_3():
    b,pb = SkewedSearchSpaceIterator_args_1()

    bi = [0,1,1,0,0]
    sp = np.array([b[i,bi_] for (i,bi_) in enumerate(bi)])
    return search_space_iterator.SkewedSearchSpaceIterator(b,pb,sp,None,SSIHop = 3, cycleOn = True, cycleIs = 1)



########



from morebs2 import aprng_gauge,search_space_iterator
import numpy as np
import unittest

"""
python -m morebs2.tests.aprng_gauge_test  
"""

def equal_BatchIncrStruct_output(bis,wanted):
    stat = True 
    coll = []
    while stat:
        p = next(bis)
        stat = not (type(p) == type(None))
        if stat: coll.append(p)
    coll = np.array(coll) 
    return np.equal(coll,wanted).all() 

class TestBatchIncrStructMethods(unittest.TestCase):

    def test__BatchIncrStruct_argseq_1(self):
        bis = aprng_gauge.BatchIncrStruct(3,True,True,2)
        ans = [[0, 0],\
        [0, 1],\
        [0, 2],\
        [1, 0],\
        [1, 1],\
        [1, 2],\
        [2, 0],\
        [2, 1],\
        [2, 2]]
        assert equal_BatchIncrStruct_output(bis,ans)

    def test__BatchIncrStruct_argseq_2(self):
        bis = aprng_gauge.BatchIncrStruct(3,True,False,2)
        ans = [[0, 1],\
        [0, 2],\
        [1, 0],\
        [1, 2],\
        [2, 0],\
        [2, 1]]
        assert equal_BatchIncrStruct_output(bis,ans)

    def test__BatchIncrStruct_argseq_3(self):
        bis = aprng_gauge.BatchIncrStruct(3,False,True,2)
        ans =[[0, 0],\
        [0, 1],\
        [0, 2],\
        [1, 1],\
        [1, 2],\
        [2, 2]]
        assert equal_BatchIncrStruct_output(bis,ans)

    def test__BatchIncrStruct_argseq_4(self):
        bis = aprng_gauge.BatchIncrStruct(3,False,False,2)
        ans = [[0, 1],\
        [0, 2],\
        [1, 2]]
        assert equal_BatchIncrStruct_output(bis,ans)

class TestAPRNGGaugeMethods(unittest.TestCase):

    def test__coverage_of_sequence(self):
        vf = [0.0,3.0,7.0]
        rv = [0.0,10.0]
        max_radius = 1.0
        cov = aprng_gauge.coverage_of_sequence(vf,rv,max_radius)
        assert cov == 0.5

        max_radius = 2.0
        cov = aprng_gauge.coverage_of_sequence(vf,rv,max_radius)
        assert cov == 0.9

        max_radius = 1.5
        cov = aprng_gauge.coverage_of_sequence(vf,rv,max_radius)
        assert cov == 0.75

    def test__normalized_uwpd(self):
        v = np.array([0,1,5,-10,12])
        rv = [-10,20]
        q0 = aprng_gauge.uwpd(v,accum_op = np.add)
        ans = sum([1, 5, 10, 12, 4, 11, 11, 15, 7, 22])
        assert q0 == ans

        q1 = aprng_gauge.max_float_uwpd(5,rv)
        nuwpd = aprng_gauge.normalized_float_uwpd(v,rv)
        assert np.round(ans / q1,5) == \
            nuwpd, "got {} wanted {}".format(nuwpd,ans/q1) 

    def test__to_noncontiguous_ranges(self):
        ranges = [[0,1],[0.7,1.5],[1.6,2.0],[1.65,2.3]]
        tnr = aprng_gauge.to_noncontiguous_ranges(ranges)
        assert tnr == [[0, 1.5], [1.6, 2.3]]

    def test_APRNGGauge__measure_cycle(self):

        bounds = np.array([[0,5],\
                        [0,5],\
                        [0,5]])

        start_point = np.zeros((3,))
        column_order = [2,1,0] # [0,1,2]
        ssi_hop = 5
        cycle_on = False
        cycle_is = 0

        ssi = search_space_iterator.SearchSpaceIterator(bounds,start_point,column_order,\
            ssi_hop,cycle_on,cycle_is)
        def fq():
            return np.sum(next(ssi)) 

        ap = aprng_gauge.APRNGGauge(fq,[0,15.],0.5) 
        term_func = lambda l,l2: not (l == l2).all() if type(l2) != type(None) else True 
        mc = ap.measure_cycle(term_func=term_func)
        assert mc == [0.83333, 0.18622]
        return

if __name__ == '__main__':
    unittest.main()


'''
functions to aid in Delineation

make-shift
'''
import numpy as np

def eliminate_duplicate_points(edges1,edges2,axis):

    delete1,delete2 = [],[]
    m1,m2 = np.mean(edges1,axis=0), np.mean(edges2,axis=0)
    el1,el2 = len(edges1),len(edges2)

    for (i,x) in enumerate(edges1):
        ss = np.where((edges2 == x).all(axis=1))[0]
        if len(ss) == 0:
            continue
        assignIt = assign_duplicate_point_to_edge(m1,m2,el1,el2,axis,x)
        
        # case: assign duplicate edge 
        if assignIt == 1:
            delete2.extend(ss)
        else:
            delete1.append(i)
    
    delete2 = set(delete2)
    delete1 = set(delete1)

    delete2 = list(delete2 - {0,el2 - 1})
    delete1 = list(delete1 - {0,el1 - 1})

    edges1 = np.delete(edges1,delete1,axis=0)
    edges2 = np.delete(edges2,delete2,axis=0)
    return edges1,edges2

def assign_duplicate_point_to_edge(edgesMean1,edgesMean2,edges1Len,edges2Len,axis,p):
    x1 = abs(edgesMean1[axis] - p[axis]) / edges1Len
    x2 = abs(edgesMean2[axis] - p[axis]) / edges2Len

    if x1 < x2: return 1
    return 2


######################### approach: no cross

def eliminate_contrary_points(e1,e2,axis):
    '''
    * assume the point sequences e1 and e2 are
    ordered by axis 0 or 1. 


    if duplicate: assign t
    '''

    # use the last element as first point to edges
    e1_,e2_ = [e1[0]],[e2[0]]
    e1,e2 = e1[1:],e2[1:]

    while len(e1) > 1 and len(e2) > 1:
        edge1 = np.array([e1_[-1],e1[0]])
        
        # iterate through e2 and pop each point not in range
        # stopping at the first point
        j = next_point_in_range(e2,edge1,axis)
        if j == -1:
            e1_.append(e1[0])
            e1 = e1[1:]
            continue

        for j2 in range(j):
            e2_.append(e2[0])
            e2 = e2[1:]

        # check edge 1 and edge 2
        edge2 = np.array([e2_[-1],e2[0]])
        s1,s2 = correct_rectangle_cross(edge1,edge2,axis)

        # case: delete the second point
        if not s1:
            e1 = e1[1:]

        if not s2:
            e2 = e2[1:]
        else:
            e2_.append(e2[0])
            e2 = e2[1:]
    e1_.extend(e1)
    e2_.extend(e2)
    return e1_,e2_

def eliminate_contrary_points_indices(e1,e2,axis):
    e1_,e2_ = [0],[0]
    ei1,ei2 = 1,1
    l1,l2 = len(e1) - 1, len(e2) - 1
    while ei1 < l1 and ei2 < l2: 
        edge1 = np.array([e1[e1_[-1]],e1[ei1]])
        j = next_point_in_range(e2[ei2:],edge1,axis)

        if j == -1:
            e1_.append(ei1)
            ei1 += 1
            continue

        for j2 in range(j):
            e2_.append(j2 + ei2) 

        edge2 = np.array([e2[e2_[-1]],e2[j + ei2]])
        s1,s2 = correct_rectangle_cross(edge1,edge2,axis)
        
        # case: edges cross for s1
        if not s1:
            ei1 += 1 

        # case: edges do not cross for s2
        if s2:
            e2_.append(ei2) 
        ei2 += j + 1

    e1_.extend([i for i in range(ei1,l1 + 1)])
    e2_.extend([i for i in range(ei2,l2 + 1)])
    return e1_,e2_

def next_point_in_range(ep,edge,axis):

    for (i,p) in enumerate(ep):
        if p[axis] >= min(edge[:,axis])\
            and p[axis] <= max(edge[:,axis]):
            return i
    return -1

def next_point_that_crosses_diag_area(ep,d,axis):
    '''
    # NOTE: inefficient
    
    return := index of next point of `ep` that is in diagonal `d`
    '''
    for (i,p) in enumerate(ep):
        if point_in_diag_area(p,d): return i
    return -1


def correct_rectangle_cross(l1,l2,axis):
    q = line_segments_cross_in_rectangle(l1,l2)
    p1,p2 = True,True
    
    if q[0][1]:
        if closer_edge(l1[0],l2[0],l1[1],axis) != 1:
            p1 = False

    if q[1][1]:
        if closer_edge(l1[0],l2[0],l2[1],axis) != 2:
            p2 = False
            
    '''
    if q[0][1]:
        p1 = False
    if q[1][1]:
        p2 = False
    '''
    return p1,p2


def closer_edge(e1r,e2r,p,axis):
    a2 = (axis + 1) % 2
    d = abs(e1r[a2] - p[a2])
    d2 = abs(e2r[a2] - p[a2])

    if d < d2:
        return 1
    return 2

def line_segments_cross_in_rectangle(l1,l2):
    '''
    determines if l1[1] or l1[0] in rectangle containing l2
    or vice-versa.

    return := sequence<(0|1 denoting line,point 0|1 of line)> of points
              that fall in rectangles.
    '''

    s10 = point_in_diag_area(l1[0],l2)
    s11 = point_in_diag_area(l1[1],l2)
    s20 = point_in_diag_area(l2[0],l1)
    s21 = point_in_diag_area(l2[1],l1)

    s2 = point_in_diag_area(l2[1],l1)
    return ((s10,s11),(s20,s21))

def point_in_diag_area(p,d):
    return p[0] >= min(d[:,0]) and p[0] <= max(d[:,0])\
        and p[1] >= min(d[:,1]) and p[1] <= max(d[:,1])

######################## approach: no jags
def remove_jags_on_edges(edges,direction):
    je = jags_on_edges(edges,direction)
    je.pop(-1)
    indices = []
    for (i,j) in enumerate(je):
        if j == -1:
            indices.append(i + 1)
    indices = np.array(indices)
    if len(indices) != 0: 
        edges = np.delete(edges,indices,0)
    return edges

def jags_on_edges(edges,direction):
    '''
    a jag is defined as three consecutive
    points in a sequence of points ordered
    by an axis q such that the middle point
    is directionally less than the other two
    points, thereby creating a triangular shape
    without the bottom edge. 
    '''

    x = edges[0]
    edges = edges[1:]
    jags = []
    while len(edges) > 0:
        q = np.array([x,edges[0]])
        ij = is_jagged(q,direction)
        if ij:
            jags.append(-1)
        else:
            jags.append(1)
        x = edges[0]
        edges = edges[1:]
    return jags

def is_jagged(edge,direction):

    if direction == 'l':
        return edge[1,0] > edge[0,0]
    
    if direction == 'r':
        return edge[1,0] < edge[0,0]

    if direction == 't':
        return edge[1,1] < edge[0,1]

    if direction == 'b':
        return edge[1,1] > edge[0,1]

############## approach: partitioning

def critical_dpoints_by_partition(edges,direction,p):

    if direction == 'l':
        qf = lambda em: np.argmin(em[:,0])
    elif direction == 'r':
        qf = lambda em: np.argmax(em[:,0])
    elif direction == 't':
        qf = lambda em: np.argmax(em[:,1])
    else:
        qf = lambda em: np.argmin(em[:,1])

    i = 1
    psz = len(edges) / p
    ps = [np.copy(edges[0])]
    while True:
        se = edges[i:i + psz]
        if len(se) == 0: break

        si = qf(se)

        ps.append(np.copy(se[si]))
        i += psz
    
    ps.append(np.copy(edges[-1]))
    return np.array(ps)

###################################################

def delineation_in_delineation(d1,d2):

    # go curve by curve
    



    return -1

# given i1,i2 for e1,e2
# 
#
# if 


from .poly_interpolation import *
from .numerical_extras import fractional_add,fractional_mul,reduce_fraction
from copy import deepcopy
import random

def merge_duplicates_cevector(v):
    """merges elements of a (coefficient,exponent) vector
    so that there are no duplicate exponents
    """
    d = {}
    for v_ in v:
        x = v_[0]
        if v_[1] in d:
            x += d[v_[1]]
            if x == 0:
                del d[v_[1]]
            else:
                d[v_[1]] = x
        else:
            d[v_[1]] = x
    return d

def merge_duplicates_ceqvector(v):
    d = {}
    for v_ in v:
        x1 = deepcopy(v_) 
        if v_[2] in d:
            x1 = (deepcopy(d[v_[2]]), deepcopy(v_[:2]))
            x1 = fractional_add(x1[0],x1[1])
            if x1[0] == 0:
                del d[v_[2]]
                continue
        d[v_[2]] = x1  
    return d

class CEQPoly:

    def __init__(self,v):
        if len(v) > 0:
            assert v.shape[1] == 3, "invalid"
        assert len(np.unique(v[:,2])) == len(v), "duplicate exponent in c-e format"
        indexOrder = np.argsort(v[:,2])[::-1]
        v = v[indexOrder]
        self.v = []
        for v_ in v:
            if v_[0] == 0: continue
            self.v.append(v_)
        self.v = np.array(self.v)
    
    def __str__(self):
        s = ""
        for y_ in self.v:
            s += " +" if y_[0] >= 0 else " -"
            c = "(" + str(abs(y_[0])) + "/" + str(y_[1]) + ")"
            s += " {}x^{}".format(c,y_[2])
        if len(s) == 0:
            return s
        return s[2:]

    def apply(self,x):
        s = 0.0
        for v_ in self.v:
            s += (v_[0] /v_[1] * x ** v_[1])
        return s

    def __add__(self,s):
        v1 = deepcopy(self.v)
        v2 = deepcopy(s.v)

        # case: one of the additives is 0.
        if len(v1) == 0:
            return deepcopy(s)
        if len(v2) == 0:
            return deepcopy(self)
        v2 = list(v2)
        v3 = []

        for v in v1:
            index = -1
            for (i,vx) in enumerate(v2):
                if round(v[2],5) == round(vx[2],5):
                    index = i
                    break

            if index == -1:
                v3.append(v)
            else:
                s1,s2 = fractional_add((v[0],v[1]),(v2[0],v2[1]))
                v3.append((s1,s2,v[2]))
                v2.pop(index) 

        # add the remainder of v2
        for v in v2:
            v3.append(v)
        return CEQPoly(np.array(v3))

    def __mul__(self,s):
        if type(s) in {int,float,complex,tuple}:
            return self.mul_s(s)

        ecv = []
        for v_ in self.v:
            for s_ in s.v:
                m = fractional_mul((v_[0],v_[1]),(s_[0],s_[1]))
                ecv.append((m[0],m[1],v_[2] + s_[2]))
        d = merge_duplicates_ceqvector(ecv)
        return CEQPoly.from_dict(d)

    # TODO: caution with imaginary numbers.
    def mul_s(self,s):
        v2 = []

        for v_ in self.v:
            if type(s) == tuple:
                m = fractional_mul((v_[0],v_[1]),(s[0],s[1]))
            else:
                m = reduce_fraction((v_[0] * s,v_[1]))
            v2.append((m[0],m[1],v_[2]))
        return CEQPoly(np.array(v2))

    @staticmethod
    def from_dict(d):
        '''
        dictionary is in e-c format.
        '''
        x = np.array([(v[0],v[1],k) for k,v in d.items()])
        return CEQPoly(x)

class CEPoly:
    """polynomial represented in coefficient-exponent form. 
    
    :param v: vector of values, length l -1 is power, index 0 is greatest power
    """
    
    def __init__(self,v):
        if len(v) > 0:
            assert v.shape[1] == 2, "invalid"
        assert len(np.unique(v[:,1])) == len(v), "duplicate exponent in c-e format"
        indexOrder = np.argsort(v[:,1])[::-1]
        v = v[indexOrder]
        self.v = []
        for v_ in v:
            if v_[0] == 0: continue
            self.v.append(v_)
        self.v = np.array(self.v) 

    def __str__(self):
        s = ""
        for y_ in self.v:
            s += " +" if y_[0] >= 0 else " "
            c = str(y_[0]) if type(y_[0]) != complex else "(" + str(y_[0]) + ")"
            s += " {}x^{}".format(c,y_[1])
        if len(s) == 0:
            return s
        return s[2:]

    def apply(self,x):
        s = 0.0
        for v_ in self.v:
            s += (v_[0] * x ** v_[1])
        return s

    def __sub__(self,s):
        return self + (s * -1)

    def __add__(self,s):
        v1 = deepcopy(self.v)
        v2 = deepcopy(s.v)

        # case: one of the additives is 0.
        if len(v1) == 0:
            return deepcopy(s)
        if len(v2) == 0:
            return deepcopy(self)
        v2 = list(v2)
        v3 = []

        for v in v1:
            index = -1
            for (i,vx) in enumerate(v2):
                if round(v[1],5) == round(vx[1],5):
                    index = i
                    break

            if index == -1:
                v3.append(v)
            else:
                v3.append((v[0] + v2[i][0],v[1]))
                v2.pop(index) 

        # add the remainder of v2
        for v in v2:
            v3.append(v)
        return CEPoly(np.array(v3))

    '''
    singleton multiplication
    '''
    def __mul__(self,s):
        if type(s) in {int,float,complex}:
            return self.mul_s(s)

        ecv = []
        for v_ in self.v:
            for s_ in s.v:
                ecv.append((v_[0] * s_[0],v_[1] + s_[1]))
        d = merge_duplicates_cevector(ecv)
        return CEPoly.from_dict(d)

    
    def __eq__(self,s):
        return str(self - s) == "" 
    
    def mul_s(self,s):
        v2 = deepcopy(self.v)
        v2 = np.array([(v_[0] * s,v_[1]) for v_ in v2])
        return CEPoly(v2)

    @staticmethod
    def from_dict(d):
        '''
        dictionary is in e-c format.
        '''
        x = np.array([(v,k) for k,v in d.items()])
        return CEPoly(x)

    def to_dict(self):
        x = {}
        for v_ in self.v:
            x[v_[1]] = v_[0]
        return x

    def c_of_p(self,p):
        '''
        coefficient of power
        '''
        for v_ in self.v:
            if round(v_[1] - p,5) == 0.:
                return v_[0]
        return None

    def exp_bounds(self):
        '''
        exponent bounds
        '''
        return (self.v[0,1],self.v[-1,1])

    def min_right_exp(self,distance = 1):
        '''
        '''
        l = len(self.v) - 1

        # case: polynomial length is 1.
        if l == 0: return None

        prev = deepcopy(self.v[0])
        for i in range(1,l):
            if prev[1] - self.v[i,1] > distance:
                return prev[1]
            prev = deepcopy(self.v[i])
        return None

    def max_left_exp(self,distance = 1):
        '''
        '''
        l = len(self.v) - 1

        # case: polynomial length is 1.
        if l == 0: return None

        prev = deepcopy(self.v[l])
        for i in range(l - 1,-1,-1):
            if prev[1] + distance < self.v[i,1]:
                return prev[1]
            prev = deepcopy(self.v[i])        
        return None

    def exponential_subtract(self,p2):
        return set(self.v[:,1]) - set(p2.v[:,1])

class SPoly:
    """polynomial operator over 1 float variable;
    data structure uses vector-index notation.

    :param v: vector of values, length l -1 is power, index 0 is greatest power
    """

    def __init__(self,v):
        assert is_vector(v), "invalid vector"

        i = 0
        for (j,v_) in enumerate(v):
            if v_ != 0:
                break
            i = j + 1

        self.v = v[i:]

    def __str__(self):
        y = self.ce_notation()
        s = ""
        for y_ in y:
            s += " +" if y_[0] >= 0 else " "
            s += " {}x^{}".format(y_[0],y_[1])
        if len(s) == 0:
            return s
        return s[2:]

    def apply(self,x):
        s = 0.0
        l = len(self.v) - 1
        for v_ in self.v:
            s += (v_ * x ** l)
            l -= 1
        return s

    def __add__(self,s):
        v1 = deepcopy(self.v)
        v2 = deepcopy(s.v)

        # make the two vectors equal in length by adding 0's at
        # the beginning of the shorter one 
        l = len(v1) - len(v2)
        
        # case: add to v1
        if l < 0:
            v3 = np.zeros(-1 * l)
            v1 = np.hstack((v3,v1))
        else:
            v3 = np.zeros(l)
            v2 = np.hstack((v3,v2))
        return SPoly(v1 + v2)

    def __sub__(self,s):
        return self + (s * -1)

    def __mul__(self,s):
        if type(s) in [int,float,complex]:
            return SPoly(s * self.v) 

        l2 = len(self.v) -1 + len(s.v)
        v = np.zeros(l2)
        v2 = np.zeros(l2)

        le1 = len(self.v) - 1
        le2 = len(s.v) - 1

        for (i,x) in enumerate(self.v):
            v3 = deepcopy(v2)
            e1 = le1 - i
            for (j,x2) in enumerate(s.v):
                x3 = x * x2
                e2 = le2 - j
                e3 = e1 + e2
                v3[l2 - e3 - 1] = x3
            v = v + v3
        return SPoly(v) 

    def vector_index_notation(self):
        return deepcopy(self.v)

    def ce_notation(self):
        '''
        coefficient-exponent notation
        '''
        p = []
        l = len(self.v) - 1 

        for (i,v_) in enumerate(self.v):
            if v_ == 0.:
                continue
            p.append((v_,l - i))
        return np.array(p)

    @staticmethod
    def from_ce_notation(v):
        # get the max exponent
        q = max([v_[1] for v_ in v])
        q2 = np.zeros(q + 1)

        for v_ in v:
            q2[q - v_[1] - 1] = v_[0]
        return SPoly(q2)

class ISPoly:
    """
    inversion of SPoly
    """

    def __init__(self,x):
        """
        v := vector of values, length l -1 is power, index 0 is greatest power
        """
        assert type(x) == float, "invalid x"
        self.x = x

    def apply(self,v):
        q = SPoly(v)
        return q.apply(self.x)

class RandomPolyGenerator:
    """generates a random polynomial, an expression represented
    by :class:`SPoly` that has coefficients in the range
    `coefficientRange` with a NON-ZERO 

                `cx^maxExp; c is a coefficient.`
                
    The rest of the values have a probability of `nullPr`
    for being zero.

    :param rsf: seed object, should have the two functions 
                (`randrange(min int,max int)` => int) and (`random()` => x in [0,1])
    :type rsf: seed-like object
    :param coefficientRange:
    :type coefficientRange:
    :param maxExp:
    :type maxExp:
    :param nullPr:
    :type nullPr:
    """

    def __init__(self,rsf,coefficientRange,maxExp,nullPr):
        assert type(coefficientRange[0]) == int and type(coefficientRange[1]) == int, "invalid coefficient range,cond.1"
        assert coefficientRange[1] >= coefficientRange[0], "invalid coefficient range,cond.2"
        assert coefficientRange[1] != coefficientRange[0],"invalid coefficient range,cond.3"
        assert type(nullPr) == float and nullPr >= 0. and nullPr <= 1., "invalid null probability"

        self.rsf = rsf
        self.coefficientRange = coefficientRange
        self.maxExp = maxExp
        self.nullPr = nullPr

    def output(self):
        # non-zero coefficient for the first 
        p = np.zeros(self.maxExp + 1)
        x = 0

        while x == 0:
            x = self.rsf.randrange(self.coefficientRange[0],self.coefficientRange[1])
            p[0] = x
        
        l = len(p)

        # for the remainder, random chance of non-zero value
        for j in range(1,l):
            if self.rsf.random() > self.nullPr:
                x = 0
                while x == 0:
                    x = self.rsf.randrange(self.coefficientRange[0],self.coefficientRange[1])
                    p[j] = x
        return SPoly(p)
        


from .poly_interpolation import *


class ChainedLagrangePolySolver:
    '''
    extended version of :class:`LagrangePolySolver` that interpolates points
    over n dimensions.

    * Point-set conditions for use:
    
    - non-duplicate values for each dimension
    
    * Requires a valid axis order of interpolation:
    
    - path that allows for lagrange interpolation in n-dim.
    '''

    def __init__(self, points):
        self.points = points
        self.assert_valid_points()
        self.oi = None
        self.lapassio = None
        return

    '''
    '''
    def assert_valid_points(self):
        assert is_2dmatrix(self.points), "invalid points"
        for i in range(self.points.shape[1]):
            q = self.points[:,i]
            assert len(set(q)) > 2, "invalidati invalidatini #2"

    def set_axis_order_of_interpolation(self,o):
        assert len(set(o)) == len(o)
        assert max(o) == self.points.shape[1] - 1
        assert min(o) == 0
        self.oi = o

    def interpolate_pair(self,ref,on):
        refs = self.points[:,ref]
        ons = self.points[:,on]
        pts = np.array([refs,ons]).T
        lps = LagrangePolySolver(pts)
        return lps#.form_point_sequence(hopIncrement = hopIncrement,capture = False)

    """
    """
    def formulate_ordering(self):
        q = [i for i in range(self.points.shape[1])]

        def swap_pathos(p,i,j):
            p[j],p[i] = p[i],p[j]

        pathos = np.copy(self.oi)
        # if i does not end at the second to last...
        for i in range(self.points.shape[1] - 1):
            # check i for swap
            q = np.copy(self.points[:,i])
            q2 = np.copy(self.points[:,i+1])
            if len(set(q)) < 2:
                print("swapping index pair")
                swap_pathos(pathos,i,i+1)
        return pathos

    # CAUTION: unstable formulation
    def check_valid_ordering(self):
        noi = self.formulate_ordering()
        self.oi = noi

        # check the last 2 indices
        x1,x2 = self.oi[-1],self.oi[-2]
        q1,q2 = self.points[:,x1],self.points[:,x2]

        return len(set(q1)) == len(set(q2))

    '''
    sets a lapassio, a sequence of LagrangePolySolvers
    '''
    def formulate(self):
        assert self.check_valid_ordering(),"invalid ordering"

        self.lapassio = []

        res = self.points[:,0].reshape((self.points.shape[0],1))
        for i in range(self.oi.shape[0] - 1):
            q = self.interpolate_pair(i,i+1)#[:,1]
            self.lapassio.append(q)
        return

    '''
    '''
    def at_x1(self,x1):
        assert x1 >= min(self.points[:,self.oi[0]]) and\
            x1 <= max(self.points[:,self.oi[0]])

        p = [x1]
        for i in range(self.points.shape[1] - 1):
            q = self.lapassio[i].output_by_lagrange_basis(p[-1])
            p.append(q)
        return np.array(p)

    def form_point_sequence(self,hopIncrement):
        assert type(self.oi) != type(None), "invalidati invalidatini"
        res = self.points[:,0].reshape((self.points.shape[0],1))
        for i in range(self.oi.shape - 1):
            q = self.interpolate_pair(i,i+1,hopIncrement)
            q = q.form_point_sequence(hopIncrement = hopIncrement,capture = False)[:,1]
            res = np.vstack((res,q))
        return res



from .ball_comp_components import *

# CAUTION: modifies arg<ball> and may results in misallocated variables
# TODO: use `savePointCopy`
# TODO: case: ball falls under (literal,r)
class BallOperator:

    def __init__(self, ball,savePointCopy = False):
        """
        an operator that can:
        - navigate balls by the "division"-scheme
        - split balls into sub-balls


        # Ball splitting procedures
        - split value
        * literal: float value, the radius of the subball
        * dividor: float value, divides the distance between the target point of the sub-ball
                                and the ball center

        - sub-ball radius
        * static: sub-ball will attempt to fulfill radius value requirement
        * minimal: algorithm will use sub-ball's actual radius, after the
                   method <add_ball_points_to_subball>

        :param ball: instantia ball
        :type ball: ball
        """
        self.ball = ball
        self.savePointCopy = savePointCopy
        self.clear_cache()

        # variables for navigation
        self.location = None
        self.basis = None
        self.nHop = None
        self.counterLocation = None

        # float value, floor corresponds to division
        self.divMarker = 0.0

        ## used for ball-splitting operations
        self.subballs = []
        self.subballRadii = []

        # target point is running radial reference of ball's points
        # not in sub-ball
        self.targetPoint = None
        return

    def clear_cache(self):
        self.cache = np.empty((0,self.ball.center.shape[0]))
        self.cache2 = None
        return


    ####################### TODO:
    ##################################### start: navigation

    # TODO: add arg<ordered>
    @staticmethod
    def nav_basis(point):
        """
        creates de basista navigatadas de trabadors aus swndtin.

        :param point: vectadas
        :type point: np.ndarray
        :return: basis
        :rtype: np.ndarray
        """
        assert is_vector(point), "invalid point"
        x = [point]
        for i in range(1,point.shape[0]):
            x2 = point[:point.shape[0] - i]
            x1 = point[point.shape[0] - i:]
            x_ = np.append(x1,x2)
            x.append(x_)
        return np.array(x)

    @staticmethod
    def calculate_counterpoint(point,center):
        """
    a counterpoint is the point of equal distance to ||`point` - `center`||
    on the line containing the points, `point` and `center`.

        :param point: vectadas
        :type point: np.ndarray
        :param center: vectadas
        :type center: np.ndarray

        :return: counterius de vectadas
        :rtype: np.ndarray
        """
        d = point - center
        return center - d

    def set_navigation(self,p):
        """
        sets basista

        :param
        """
        basis = BallOperator.nav_basis(p)
        self.location = p
        self.bGen = generate_possible_binary_sequences(p.shape[0], [], elements = [1,-1])
        self.counterLocation = BallOperator.calculate_counterpoint(self.location,self.ball.center)

    """
    """
    def nav_one(self,operator):
        return -1


    """
    """
    @staticmethod
    def basis_division(point,counterpoint,multiplier):
        assert is_vector(point) and is_vector(counterpoint), "invalid args. point and center"

        x = []
        for (i,p) in enumerate(point):
            if multiplier[i] == 1: x.append(p)
            else: x.append(counterpoint[i])
        x = np.array(x)
        return BallOperator.nav_basis(x)

    @staticmethod
    def hop_in_division(b, hop, center):
        assert len(b.shape) == 2 and b.shape[0] == b.shape[1], "incorrect shape for basis"
        assert hop >= 0.0 and hop <= 1.0, "invalid hop"

        i = int(math.floor(hop / (1/b.shape[0])))
        i2 = (i + 1) % b.shape[0]
        q = np.vstack((center,b))
        r = hop - i * 1/b.shape[0]
        c = r / (1 / b.shape[0])
        return q[i] + c * (q[i2] - q[i])

    ##################################### end: navigation

    def run_subball_split(self, split,subballRadiusType = "static", verbose = False):
        """
        Continually splits the ball into sub-balls until all points lie w/in 1+ sub-ball.

        :param split: "literal","dividor"
        :type split: str
        :param subballRadiusType: "static","minimal"
        :type subballRadiusType: str
        :param verbose: verbosicas?
        :type subballRadiusType: bool
        """
        assert split[1] in {"literal","dividor"}, "invalid split"
        assert subballRadiusType in {"static","minimal"}, "invalid subball radius type"

        self.subballs = []
        self.subballRadii = []

        if verbose:
            print("\t\t\tstarting split on ball of size {}".format(self.ball.data.newData.shape[0]))


        if split[1] == "literal" and self.ball.radius <= split[0]:
            self.subballs.append(deepcopy(self.ball))
            self.ball.data.newData = np.empty((0,self.ball.data.newData.shape[1]))
            return

        while self.ball.data.newData.shape[0] > 0:
            self.update_radial_ref()
            if split[1] == "dividor":
                sr = cr(euclidean_point_distance(self.ball.center,\
                    self.targetPoint)  / split[0])
            else:
                sr = cr(split[0])

            sb = self.subball_at_radial_ref(sr,self.targetPoint)
            self.subballs.append(sb)
            q = sr if subballRadiusType == "static" else sb.radius
            self.subballRadii.append(q)

            self.cache2 = np.copy(sb.data.newData)

            print("\t\t\tnumber of subballs: {} | added points {}".format(len(self.subballs), sb.data.newData.shape[0]))

            self.add_cache_points_to_subball(sb,q)
            self.cache = np.vstack((self.cache,self.cache2))
        return

    def revert_split(self):
        """
        thee reversia
        """

        self.ball.data = PointSorter(self.cache)
        self.clear_cache()
        while len(self.subballs) > 0:
            x = self.subballs.pop(0)
            del x
        self.subballs = []

    def antiradial_ref(self,subballRadius, radialRef):
        """
        antiradinox reference de la pointis

        :return: pointis de vectadas
        :rtype: np.ndarray
        """
        i = np.argmin([abs(euclidean_point_distance(p,radialRef) - subballRadius) for p in self.ball.data.newData])
        return self.ball.data.newData[i]

    def subball_at_radial_ref(self, subballRadius, radialRef):
        """
        draw line segment (radialRef,e) towards ball center of length
        `subballRadius`; e is the center

        :param subballRadius:
        :type subballRadius:
        :param radialRef:
        :type radialRef:
        :return: ball instance
        :rtype: Ball
        """

        #### new calculation for center
        ref2 = self.antiradial_ref(subballRadius,radialRef)
        newCenter = ref2
        b = Ball(newCenter)
        subballRadius = cr(euclidean_point_distance(b.center,radialRef)) + 10 ** -4

        # save remaining target ball points to subball
        self.add_ball_points_to_subball(b,subballRadius)
        return b

    ############## start: method for after post-add

    def add_cache_points_to_subball(self,sb,r):
        """
        adds each point p already iterated over (in cache) to subball `sb`
        if p is at distance <= r to `sb.center`

        :param sb: sobballco
        :type sb: Ball
        :param r: floatist
        :type r:  float
        """
        for x in self.cache:
            if euclidean_point_distance(sb.center,x) <= r:
                sb.add_element(x)

    ############## end: method for after post-add

    def update_radial_ref(self):
        index = np.argmax([euclidean_point_distance(self.ball.center,x) for x in self.ball.data.newData])
        self.targetPoint = self.ball.data.newData[index]

    def add_ball_points_to_subball(self,subball, subballRadius):
        BallOperator.ball_points_to_another(self.ball,subball,subballRadius,True)

    @staticmethod
    def ball_points_to_another(b1,b2,b2Radius,deleteFromB1):
        """
    Transfers each point p from .`b1` to `b2` if p falls w/in distance of
    `b2Radius`. If `deleteFromB1` is set to True, delete p from .`b1` if `b2`
    accepts p.
        """

        ### case:
        # farthest possible point in b1 with respect to b2 falls out of b2Radius
        ed = cr(euclidean_point_distance(b1.center,b2.center))
        if ed - b1.radius > b2Radius:
            return

        i = 0
        while i < b1.data.newData.shape[0]:
            x = b1.data.newData[i]
            ed = cr(euclidean_point_distance(x,b2.center))
            if ed <= b2Radius:
                b2.add_element(x)

                if deleteFromB1:
                    b1.data.newData = np.concatenate(\
                        (b1.data.newData[:i],b1.data.newData[i + 1:]))
                else:
                    i += 1
            else:
                i += 1
        return



from scipy import optimize as opt
from .random_generators import *
from .travel_data import *
import logging

powerRow = lambda x, l: np.append(np.array([x ** (l - i) for i in range(l)]), np.array([1]))

# the effect of gravity at any instant, used to calculate acceleration
GRAVITY_AT_INSTANT = 11

class LagrangePolySolver:
    '''
    basic polynomial solver by Lagrange basis functions

    polynomial solver class that finds an interpolating function for a set of
    2-dimensional points.

    :class:`LagrangePolySolver` also has the capability to calculate information concerning
    the curve such as velocity data (using an initial velocity) by the gravity variable `GRAVITY_AT_INSTANT`,
    in which it collates that information into :class:`TravelData` instances.

    :param points: sequence of n-dimensional points
    :param prefetch: perform pre-processing (x-bounds and integral length) on points?
    :type prefetch: bool
    '''


    def __init__(self, points, prefetch = True):
        self.ps = points
        self.targetRangePoints = None
        self.vfs = None

        self.A = None

        self.recentRangeSearchInfo = None
        self.minumum, self.maximum = None,None
        self.integrl = None

        if prefetch:
            self.prefetch_data()
        self.timeActionStamp = -1

        self.allTravelData = None # type TravelData
            # variables used to declare allTravelData
        self.velocityData = []
        self.durationData = []

    #
    @staticmethod
    def yield_lines_clockwise_from_source_in_area(sourcePoint, area, numParts):
        assert is_valid_area(area), "invalid area"

        # check that source point is a corner of area
        q = quadrant_of_corner_point_in_area(sourcePoint, area)
        if q == 2:
            dRange = [0,90]
        elif q == 3:
            dRange = [90,180]
        elif q == 0:
            dRange = [180,270]
        else:
            dRange = [270,360]

        degDelta = 90.0 / (numParts + 1)
        xLength = area[1,0] - area[0,1]
        yLength = area[1,1] - area[0,1]

        maxLength = point_distance(area[0],area[1])
        for i in range(numParts):
            deg = dRange[0] + degDelta * (i + 1)
            l = generate_line_by_length_and_angle(sourcePoint, maxLength, deg)
            yield l.trim_to_fit_in_area(area)

    def prefetch_data(self):
        self.bounds_for_x("float")
        self.integral_length(self.minumum, self.maximum, integrationHop = DEFAULT_TRAVELLING_HOP)

    def display_basic(self):
        print("\t* min: ", self.minumum, "\t* max: ", self.maximum)
        print("\t* points: ")
        print(self.ps)
        print("\t* target range points: ")
        print(self.targetRangePoints)
        print("\t* integral: ", self.integrl)

    ########################################## START: lagrange basis

    """
    TODO: opt., make lambda function instead??
    """
    def lagrange_basis(self, x, pointIndex):
        v = 1.0

        for i in range(len(self.ps)):
            if i == pointIndex:
                continue

            num = x - self.ps[i][0]
            den = self.ps[pointIndex][0] - self.ps[i][0]

            if den == 0:
                #return 0
                print("NOT GOOD: ", self.ps[pointIndex][0], " X2 ", self.ps[i][0])
                #v = 0

            v = v * zero_div(float(num),float(den),1.0)
        return v

    def bounds_for_x(self, returnType):
        self.minumum, self.maximum = np.min(self.ps[:, 0]), np.max(self.ps[:,0])
        if returnType == "float":
            return self.minumum, self.maximum

        return int(math.ceil(self.minumum)), int(math.floor(self.maximum))

    def output_by_lagrange_basis(self, x):
        # CASE: x is 0-length
        ##if self.

        y = 0
        for i in range(len(self.ps)):
            y += self.ps[i][1] * self.lagrange_basis(x, i)
        return y

    # TODO: this needs to be fixed for missed hop at end
    def output_range(self, rangeX, intRound = False):

        self.targetRangePoints = np.empty((0,2))
        i = rangeX[0]
        xs = [y[0] for y in self.ps]
        while i < rangeX[1]:
            p = np.array([[i,self.output_by_lagrange_basis(i)]])
            self.targetRangePoints = np.append(self.targetRangePoints, p, axis = 0)
            i += 1

        if intRound:
            self.targetRangePoints = np.round(self.targetRangePoints)
            self.targetRangePoints = self.targetRangePoints.astype(int)

    ########################################## END: lagrange basis

    def x_point_exists(self, x):
        for (i,q) in enumerate(self.ps):
            if q[0] == x:
                return i
        return -1

    # TODO: code and refactor, multi-processing
    def integral_length(self, startX, endX, integrationHop = DEFAULT_TRAVELLING_HOP):
        # TODO: off-by-1 bug here
        assert integrationHop > 0, "invalid integration hop"
        integrationHop = integrationHop if startX < endX else -integrationHop
        self.integrl = None
        prevPoint = (startX, self.output_by_lagrange_basis(startX))

        integrationHop
        tFunc = lambda x: True if x <= endX and x >= startX else False

        d = 0
        while tFunc(startX):
            startX += integrationHop
            point = (startX, self.output_by_lagrange_basis(startX))
            d += point_distance(prevPoint, point)
            prevPoint = point
        self.integrl = d
        return d

    """
    return:
    - (end point)::(2-tuple)
    - (distance travelled)::(float)
    - (>= wanted distance)::(bool)
    """
    def get_xrange_for_wanted_distance(self, startPoint, distance, integrationHop = DEFAULT_TRAVELLING_HOP):

        assert not (distance <= 0), "invalid wanted distance"

        pf = lambda x: True if x >= self.minumum and x <= self.maximum else False
        df = lambda x: True if x >= distance else False
        r, q, gw = self.traversal_loop(startPoint, integrationHop, pf, df)

        # trim leftovers
        if gw:
            newHop = integrationHop / 10.0
            excess = q - distance
            prevPoint = (r[0] - integrationHop, self.output_by_lagrange_basis(r[0] - integrationHop))
            dPrev = q - point_distance(r, prevPoint)
            wanted = distance - dPrev

            df2 = lambda x: True if x >= wanted else False
            r2, q2, gw2 = self.traversal_loop(r[0] - integrationHop, newHop, pf, df2)
            return r2, dPrev + q2, gw2
        else:
            return r, q, gw

    """
    arguments:
    - startPoint: float, x-value
    - integrationHop: float, +|-
    - pFunc: function(float), point termination
    - dFunc: function(float), distance termination

    return:
    - (end point)::(2-tuple)
    - (distance travelled)::(float)
    - (>= wanted distance)::(bool)
    """
    def traversal_loop(self, startPoint, integrationHop, pFunc, dFunc):
        prevPoint, r = (startPoint, self.output_by_lagrange_basis(startPoint)),\
                    (startPoint, self.output_by_lagrange_basis(startPoint))
        totalPd = 0.0
        totalPd_ = totalPd
        gotWanted = False

        # TODO: find the cause of bug and delete fix
        noChange = 0
        c = 0
        while pFunc(prevPoint[0]):
            c += 1
            r = (prevPoint[0] + integrationHop, self.output_by_lagrange_basis(prevPoint[0] + integrationHop))
            totalPd_ = totalPd
            totalPd += point_distance(prevPoint, r)

            if dFunc(totalPd):
                gotWanted = True
                break

            if tuple(r) == tuple(prevPoint):
                noChange += 1
                if noChange >=100:
                    print("\n\n\tNO CHANGE\n\n")
                    break
            else:
                noChange = 0
                prevPoint = (r[0], r[1])

        if r[0] >= self.maximum or r[0] <= self.minumum:
            prevPoint = (prevPoint[0] - integrationHop, self.output_by_lagrange_basis(prevPoint[0] - integrationHop))
            return prevPoint, totalPd_, gotWanted
        return r, totalPd, gotWanted

    def average_slope_on_range(self, startX, endX):
        assert startX < endX, "invalid start and end x points"

        prevPoint = (startX, self.output_by_lagrange_basis(startX))
        d = 0
        # TODO: refactor this to #$#
        while startX < endX:
            startX += integrationHop
            point = (startX, self.output_by_lagrange_basis(startX))
            d += point_distance(prevPoint, point)
            prevPoint = point
        return d / float(endX - startX)

    ########################### START: vector-form solution

    def get_vfs_vars(self):
        # form A
        xs = self.ps[:,0]
        self.A = np.empty((xs.shape[0], xs.shape[0]))
        for (i,x) in enumerate(xs):
            q = powerRow(x, xs.shape[0] - 1)
            self.A[i] = q

        # form b
        ys = self.ps[:,1]
        return self.A, ys

    def vector_form_solution(self):
        self.A, ys = self.get_vfs_vars()
        self.vfs = np.linalg.solve(self.A, ys)

    def output_by_vector_form(self, x):
        q = powerRow(x, self.ps.shape[0] - 1)
        return q.dot(self.vfs)

    def distance_between_points(self, x1, x2):

        p1, p2 = self.output_by_lagrange_basis()
        return point_distance(p1, p2)

    """
    description:
    - draws a line between two polynomial points and determines the minumum velocity to travel.

      CAUTION: does not consider points in (min(x1,x2), max(x1,x2))

    arguments:
    - x1: float
    - x2: float

    return:
    -
    """
    def required_velocity_for_travel_of_range_angle_effect(self, x1, x2):

        p1,p2 = (x1, self.output_by_lagrange_basis(x1)), (x2, self.output_by_lagrange_basis(x2))
        d = point_distance(p1, p2)
        acc = self.acceleration_between_points(p1,p2)
        if acc >= 0: return 0

        q = acc * d
        accTotal = math.sqrt(abs(q))
        duration = abs(accTotal / acc )
        return d / duration

    """
    description:
    - uses above method, considers all points in range by hop increment

    return:
    - float
    """
    def required_velocity_for_travel_of_range_angle_effect_by_hop(self, x1, x2, hop):
        assert hop > 0, "hop must be greater than 0"
        assert x1 >= self.minumum and x1 <= self.maximum, "invalid arg [0]"
        assert x2 >= self.minumum and x2 <= self.maximum, "invalid arg [1]"

        hop = -hop if x2 > x1 else hop
        q1 = x2
        q2 = x2 + hop
        tfunc = lambda x: True if x >= min(x1,x2) and x <= max(x1,x2) else False

        vmin = 0.0
        while tfunc(q2):
            p1,p2 = (q2, self.output_by_lagrange_basis(q2)), (q1, self.output_by_lagrange_basis(q1))
            vmin = self.back_hop_for_velocity(p1, p2, vmin)
            q1 = q2
            q2 = round(q2 + hop, 3)
        return vmin

    def back_hop_for_velocity(self, x1, x2, velocity):
        a = self.acceleration_between_points(x1,x2)
        b = -velocity
        c = a * point_distance(x1,x2)
        soln, stat = quadratic_formula_solve_for_reals(1.0,b,c)

        # case: pos. accel.
        if not stat:
            return velocity - a
        return max(soln)

    def is_travel_over_range_feasible(self, x1, x2, hop, velocity):
        x = self.required_velocity_for_travel_of_range_angle_effect_by_hop(x1, x2, hop)
        if velocity >= x:
            return True
        return False

    def acceleration_between_points(self, x1, x2):
        if x1[0] == x2[0] and x1[1] == x2[1]: return 0

        a = angle_between_two_points(x1, x2)
        prop = - (a / 90.0)
        return GRAVITY_AT_INSTANT * prop

    '''
    travel time == 1 if dist(x1,x2) == bv and ang(x1,x2) == 0
    '''
    def velocity_between_points_by_angle_effect(self, x1, x2, baselineVelocity):
        return baselineVelocity + self.acceleration_between_points(x1, x2)

    # TODO: needs to be tested.
    """
    description:
    -

    return:
    - (current velocity)::float
    - (able to travel)::bool
    - (duration)::float
    """
    def velocity_and_duration_between_points_by_angle_effect(self, x1, x2, baselineVelocity, duration = 0.0, c = 0):

        c += 1

        if baselineVelocity <= 0 or tuple(x1) == tuple(x2): return baselineVelocity, False, duration

        # TODO: refactor below equality statement
        if x1[0] == x2[0] and x1[1] == x2[1]:
            return baselineVelocity, True, duration
        distance = point_distance(x1,x2)

        # case: acceleration will be divided
        if distance <= baselineVelocity:
            a = self.acceleration_between_points(x1,x2) * distance / baselineVelocity
            return baselineVelocity + a, True, duration + distance / baselineVelocity
        else:
        # case: travel time b/t points > 1 unit
            x1_, d, stat = self.get_xrange_for_wanted_distance(x1[0], baselineVelocity, DEFAULT_TRAVELLING_HOP if x1[0] <= x2[0] else -DEFAULT_TRAVELLING_HOP)
            baselineVelocity += self.acceleration_between_points(x1, x1_)

            if equal_iterables(x1, x1_, roundPlaces = 5):
                return baselineVelocity, False, duration

            return self.velocity_and_duration_between_points_by_angle_effect(x1_, x2, baselineVelocity, duration + 1.0, c + 1)

    def travel_time_between_points(self, x1, x2, velocity):
        return point_distance(x1,x2) / velocity

    """
    arguments:
    - startX: float, x-value
    - duration: float, >= 0
    - baselineVelocity: float, >=
    - hop: float, small

    return:
    - TravelData, (halt restricted)::bool, (bounds reached)::bool
    """
    def travel_from_start_for_duration_under_angle_effect(self, startX, duration, baselineVelocity, hop):
        # neg
        if hop <= 0:
            term = lambda x: True if x > self.maximum else False
        # non-neg
        else:
            term = lambda x: True if x < self.minumum else False

        if baselineVelocity <= 0:
            print("\tCANNOT TRAVEL ANYMORE")
            return None, True, False

        rfunc = self.continue_travelling
        tfunc = lambda x: True if x >= duration else False
        dur = [0.0]
        distance = 0.0
        prevPoint = (startX, self.output_by_lagrange_basis(startX))
        max100 = 10
        allPoints = [prevPoint]

        # travel for duration
        endPoint, allPoints, dur, baselineVelocity1, numHops, haltRestricted, stat = self.run_travel_loop(prevPoint, allPoints, rfunc, tfunc, hop, baselineVelocity, dur, "travel time history")
        hopRanges = [((0, numHops), hop)]

        # case: out-of-bounds, cut off leftovers
        if stat or haltRestricted:
            self.timeActionStamp += 1
            return TravelData(idn = self.timeActionStamp, pointData = allPoints[:-1], durationData = dur[:-1], velocityData = baselineVelocity1[:-1], hopRanges = hopRanges), haltRestricted, stat

        # run another travel loop for remainders
        leftovers = sum(dur) - duration
        newWantedDuration = dur[-1] - leftovers

        newHop = hop / 10.0
        prevPoint = allPoints[-2]
        newTFunc = lambda x: True if x >= newWantedDuration else False
        bulkDuration = sum(dur[:-1])
        bv = baselineVelocity1[-2] if len(baselineVelocity1) >= 2 else baselineVelocity
        endPoint, allPoints2, dur2, baselineVelocity2, numHops2, haltRestricted, stat = self.run_travel_loop(prevPoint, [], rfunc, newTFunc, newHop, bv, [], "travel time history")
        hopRanges = [((0, numHops), hop), ((numHops, numHops + numHops2), newHop)]

        self.timeActionStamp += 1
        return TravelData(idn = self.timeActionStamp, pointData = np.append(allPoints[:-1], allPoints2, axis = 0), durationData = dur[:-1] + dur2,\
            velocityData = np.append(baselineVelocity1[:-1], baselineVelocity2[1:], axis = 0), hopRanges = hopRanges), haltRestricted, stat

    """
    ceil

    arguments:
    - rfunc: continue_travelling termination
    - tfunc: time threshold termination

    return:
    - (point)::2d-tuple
    - (all points)::(float | list(floats))
    - (duration)::(float | list(floats))
    - (velocity)::(float | list(floats))
    - (number of hops)::float
    - (halt restriction)::bool
    - (at end of poly.)::bool
    """
    def run_travel_loop(self, prevPoint, allPoints, rfunc, tfunc, hop, baselineVelocity:float, duration:list, returnType:str):
        distance = 0.0
        numHops = 0
        velocities = [baselineVelocity]

        point = prevPoint
        bFunc = lambda p: True if p[0] >= self.minumum and p[0] <= self.maximum else False
        endByTFunc = False
        haltRestricted = False
        ldx = len(duration)
        while bFunc(prevPoint):
            point = (prevPoint[0] + hop, self.output_by_lagrange_basis(prevPoint[0] + hop))
            ld0 = len(duration)
            duration2, baselineVelocity, distance2, haltRestricted, durSat = self.one_hop_for_travel_info(point, prevPoint, baselineVelocity, duration, distance, rfunc, tfunc, returnType)

            if haltRestricted:
                break

            duration = duration2
            distance = distance2

            allPoints.append(point)
            velocities.append(baselineVelocity)
            numHops += 1

            if tfunc(sum(duration)):
                endByTFunc = True
                break
            prevPoint = point

        return (point, allPoints[-1], sum(duration), baselineVelocity, numHops, haltRestricted, not endByTFunc)\
            if returnType == "travel time" else (point, np.array(allPoints), duration, velocities, numHops, haltRestricted, not endByTFunc)

    # TODO: refactor this and replace above
    """
    return:
    - (duration)::(float | list(floats))
    - (velocity)::float
    - (distance travelled)::float
    - (negative halt restriction)::bool
    - (duration satisfied)::bool
    """
    def one_hop_for_travel_info(self, point, prevPoint, previousVelocity, prevTravelTime, distanceTravelled, rfunc, tfunc, returnType):
        assert returnType in ("travel time history", "travel time"), "invalid return type"

        velocity, stat, t = self.velocity_and_duration_between_points_by_angle_effect(prevPoint, point, previousVelocity)

        # not able to travel
        if not stat:
            haltRestricted = True
        else:
            # ? negative halt restriction
            haltRestricted = not rfunc(t)

        if haltRestricted:
            return (sum(prevTravelTime), velocity, distanceTravelled, haltRestricted, False) if returnType == "travel time" else\
                (prevTravelTime, velocity, distanceTravelled, haltRestricted, False)

        prevTravelTime.append(t)
        distanceTravelled += point_distance(prevPoint, point)

        durSat = not tfunc(sum(prevTravelTime))
        return (sum(prevTravelTime), velocity, distanceTravelled, haltRestricted, durSat) if returnType == "travel time"\
            else (prevTravelTime, velocity, distanceTravelled, haltRestricted, durSat)

    """
    # TODO: can be used for agent decision input by query
    """
    def continue_travelling(self, travelTime, travelHistory = None, query = False):

        if query:
            raise ValueError("query mode needs to be implemented.")
        else:
            return True if travelTime >= 0 else False

    ################################## START: estimation methods

    ################################## END: estimation methods

    # TODO: caution, does not check args. start and end
    def y_extremes_in_xrange(self, start, end, hopIncrement = 0.01):
        assert hopIncrement > 0, "invalid hop increment"
        if start > end:
            start,end = end,start

        ysMin, ysMax = math.inf, - math.inf
        while start < end:
            y = self.output_by_lagrange_basis(start)

            if y > ysMax:
                ysMax = y

            if y < ysMin:
                ysMin = y

            start += hopIncrement

        return ysMin, ysMax

    """
    arguments:
    - refinement: float, [0,1] specifies the degree to partition the polynomial x-range.
    - totalArea: 2 x 2 np.array
    - radius: int, direction up and down from every point of consideration in polynomial.
    - forward: bool, specifies forward or backward direction for x-range.

    return:
    - list(area)
    """
    def capture_polynomial_as_disqualifying_areas(self, refinement, totalArea, radius = 4, forward = True):
        q = (self.maximum - self.minumum) * refinement
        s = self.minumum if forward else lps.maximum
        if not forward: q = -q
        areas = []
        while s > self.minumum and s < self.maximum:
            e = s + q

            if forward and e > self.maximum:
                e = self.maximum

            elif not forward and e < self.minumum:
                e = self.minumum

            yMin, yMax = self.y_extremes_in_xrange(s,e)

            if yMin in [np.inf, -np.inf] or yMax in [np.inf, -np.inf]:
                raise ValueError("could not capture polynomial.")

            # form the four corners to an area
            minPoint, maxPoint = np.array([min(s,e), yMin]), np.array([max(s,e), yMax])
            area = np.array([minPoint, maxPoint])
            areas.append(area)

            # update
            s = e

        return areas

    ##### START: repr methods

    """
    """
    def form_point_sequence(self, reverse = False, hopIncrement = DEFAULT_TRAVELLING_HOP, capture = True):
        assert hopIncrement > 0, "hop increment has to be +"
        start = self.minumum if not reverse else self.maximum
        hopIncrement = -hopIncrement if reverse else hopIncrement
        tFunc = lambda x: True if x >= self.minumum and x <= self.maximum else False

        allPoints = []
        while tFunc(start):
            p = [start, self.output_by_lagrange_basis(start)]
            allPoints.append(p)
            start += hopIncrement

        allPoints = np.array(allPoints)
        if capture: self.data = allPoints
        return allPoints

    def poly_to_velocity_duration_measures_in_hop_iteration(self, velocity, reverse = False, hopIncrement = DEFAULT_TRAVELLING_HOP):
        # collect velocity data
        if type(self.data) is None:
            self.form_point_sequence(reverse, hopIncrement)

        self.velocityData = []
        self.durationData = []
        for i in range(self.data.shape[0] - 1):
            velocity, stat, duration = self.velocity_and_duration_between_points_by_angle_effect(self.data[i], self.data[i + 1], velocity)
            self.velocityData.append(velocity)
            self.durationData.append(duration)

            # last element of data is no travel
            if not stat:
                self.data = self.data[:i + 2]
                break

    def poly_to_travel_data(self, velocity, travelDataIdn = 0):
        if type(self.data) is None:
            self.form_point_sequence()
        self.poly_to_velocity_duration_measures_in_hop_iteration(velocity)
        hopRanges = [((0, len(self.velocityData)), DEFAULT_TRAVELLING_HOP)]
        self.allTravelData = TravelData(travelDataIdn, self.data, self.durationData, self.velocityData, hopRanges, velocity)
        self.allTravelData.velocity_to_pointwise_acceleration_data()

    def clear_travel_data_variables(self):
        self.velocityData = []
        self.durationData = []

    # TODO: test this.
    """
    """
    def intersection_with_line(self, line):

        if self.vfs is None:
            self.vector_form_solution()

        def f(x):
            s = self.output_by_vector_form(x)
            s -= (line.slope * x + line.yint)
            return s

        try:
            r1 = opt.brentq(f, self.minumum, self.maximum) #opt.newton(f, (self.maximum - self.minumum) / 2)
            return r1, True
        except:
            return math.inf, False

    def intersection_with_line_v2(self,line):
        if self.vfs is None:
            self.vector_form_solution()

        def f(x):
            s = self.output_by_vector_form(x)
            s -= (line.slope * x + line.yint)
            return s

        # center self along opposing ends
        lps2,mid = self.center_by_endpoints(1)
        l2 = line.move_line_by_axis_shifts((0,mid))
        return lps2.intersection_with_line(l2)

    ############################################################

    # TODO: method will replace above.
    # TODO: wrong.
    def center_axis_at_zero(self, axis):
        assert axis in [0,1], "invalid axis {}".format(axis)

        if axis == 0:
            mini,maxi = self.minumum, self.maximum
        else:
            mini, maxi = self.y_extremes_in_xrange(self.minumum, self.maximum, hopIncrement = 0.05)

       # get center
        mid = (maxi - mini ) /  2.0
        dest = -mid
        reloc = - abs(mini - dest)
        newData = np.copy(self.ps)

        for i in range(newData.shape[0]):
            newData[i,axis] += reloc
        return LagrangePolySolver(newData), reloc

    def center_at_zero(self):
        q,xs = self.center_axis_at_zero(0)
        q2,ys = q.center_axis_at_zero(1)
        return q2,xs,ys

    """
    centers endpoints along axis
    0 -> balanced along y-axis
    1 -> balanced along x-axis
    """
    def center_by_endpoints(self, axis):
        assert axis in [0,1], "invalid axis"

        # calc. midpoint
        # set midpoint as 0 for axis and shift all other points
        #       by - midpoint
        mid = (self.ps[0,axis] + self.ps[1,axis]) / 2.0

        q = np.copy(self.ps)
        for i in range(q.shape[0]):
            q[i,axis] = q[i,axis] - mid
        return LagrangePolySolver(q), -mid

    ##### END: repr methods



from morebs2 import rssi
import numpy as np
import operator

def test_rch_chain_1(rchl = 1):
    """
    a version of the first chain; used in unit testing

    :param rchl: 0|1|*, cf function identifier
    :type rch1: int
    """

    rf = np.array([8.0,2.3,3.1,4.5,8.8])
    dm = rssi.euclidean_point_distance
    dt = 5.0

    if rchl == 1:
        cf = operator.gt
    elif rchl == 2:
        cf = operator.lt
    else:
        cf = lambda x, c: x + c >= 2.5 and x - c <= 5.0

    kwargs = ['r', rf,dm,cf,dt]
    rc = rssi.RChainHead()
    rc.add_node_at(kwargs)
    return rc

def sample_rssi_1(rmMode = "relevance zoom", rchl = 1):

    bounds = np.array([[-7.0,12.0],\
                        [3.0,25.0],\
                        [-20.0,-3.0],\
                        [9.0,28.0],\
                        [-2.0,32.0]])

    startPoint = np.copy(bounds[:,0])
    ssih = 2

    rc = test_rch_chain_1(rchl)
    rm = (rmMode, rc)
    return rssi.ResplattingSearchSpaceIterator(bounds, startPoint, columnOrder = None, SSIHop = ssih, resplattingMode = rm)



from .measures import *
from .globalls import *

## line generators
def generate_line_by_length_and_angle(startPoint, length, angle):
    endPoint = hypotenuse_from_point(startPoint, length, angle)
    # round startpoint and endpoint to default depth 5
    startPoint,endPoint = np.round(startPoint,5), np.round(endPoint,5)
    return Line((startPoint, endPoint))

def hypotenuse_from_point(point, length, angle):
    """
    outputs an endpoint given `point`; `endpoint` is distance `length` from .`point` and at `angle`
    """

    # get the x-delta and y-delta
    q = math.sin(math.radians(angle))
    opp = q * length

    q = math.cos(math.radians(angle))
    adj = q * length

    return [point[0] + adj, point[1] + opp]

class Line:
    """
line/line segment class
    """

    def __init__(self, endpoints, calculateCenter = True):
        endpoints = np.array(endpoints)
        assert len(endpoints.shape) == 2 and endpoints.shape[0] == 2 and endpoints.shape[0] == endpoints.shape[1], "invalid endpoints"
        assert len(endpoints) == 2, "invalid number of endpoints"
        self.endpoints = endpoints
        self.gather_variables(calculateCenter)
        self.data = None

    def __str__(self):
        return vector_to_string(self.endpoints[0], float_func) + "|" + vector_to_string(self.endpoints[1], float_func)

    def __eq__(self, l2):
        if l2 == None: return False
        return equal_iterables(self.endpoints, l2.endpoints)

    def n_partition(self,n, orderedByEnds = False):
        """
        partitions into n parts by axis with greatest change
        """
        assert n > 1 and type(n) is int, "invalid n"

        g = self.axis_with_greater_change()
        d = abs(self.endpoints[0,g] - self.endpoints[1,g])
        q = d / float(n)
        s = min([self.endpoints[0,g], self.endpoints[1,g]])

        r = np.empty((n + 1, 2))
        for i in range(n + 1):
            x = s + (q * i)
            if g == 0:
                y = self.y_given_x(x)
                r[i] = [x,y]
            else:
                y = self.x_given_y(x)
                r[i] = [y,x]

        if orderedByEnds:
            if not equal_iterables(r[0], self.endpoints[0]):
                return r[::-1]
        return r

    # TODO: not tested.
    @staticmethod
    def max_mod_distance_of_points_in_area(area, points, axisOfMod):
        assert is_valid_area(area), "invalid area"
        assert is_2dmatrix(points), "invalid points"
        assert axisOfMod in [0,1], "invalid axis of mod."

        q = np.empty((points.shape[0], 2))
        for i in range(points.shape[0]):
            assert point_in_area(tuple(points[i]), area), "point {} not in area\n{}".format(points[i],area)
            minDistance = abs(area[0,axisOfMod] - points[i,axisOfMod])
            maxDistance = abs(area[1,axisOfMod] - points[i,axisOfMod])
            q[i] = [minDistance, maxDistance]
        return q

    # TODO: check intersection on T and L shapes.
    # TODO: not tested.
    def crosses_area_edge(self,area, dualSidedCross = True):
        assert is_valid_area(area), "invalid area"

        c = area_to_corners(area)

        for i in range(4):
            l = Line(np.array([c[i], c[(i + 1) %4]]))
            p, stat = intersection_between_two_lines(l, self)
            if stat:
                if dualSidedCross:
                    if l.is_vertical():
                        if p[0] > self.endpoints[0,0] and p[0] < self.endpoints[1,0]:
                            return p,i
                    else: # h
                        if p[1] > self.endpoints[0,1] and p[1] < self.endpoints[1,1]:
                            return p,i
                else:
                    return p, i

        return None,-1

    def axis_with_greater_change(self):
        d0,d1 = abs(self.endpoints[0,0] - self.endpoints[1,0]), abs(self.endpoints[0,1] - self.endpoints[1,1])
        return 0 if d0 > d1 else 1

    def is_horizontal(self):
        return True if self.slope <= 10 ** -5\
        and abs(self.endpoints[0,1] - self.endpoints[1,1])\
        < 10 ** -4 else False

    def is_vertical(self):
        return True if self.slope <= 10 ** -5\
        and abs(self.endpoints[0,0] - self.endpoints[1,0])\
        < 10 ** -4 else False

    def change_in_xcoord_by_distance(self,distance):
        # TODO: ? sign restriction?
        assert distance >= 0, "distance cannot be negative"
        distanceDeltaPerX = math.sqrt(self.slope ** 2 + 1)
        return distance / distanceDeltaPerX


    def trim_from_end(self, endpoint, trimLength):
        """
        line segment (q, endpoint) of `trimLength` is removed
        and a new Line is returned
        """

        assert trimLength >= 0, "trim length must be pos."
        remLength = point_distance(self.endpoints[0],self.endpoints[1]) - trimLength
        if remLength <= 0:
            return None

        if equal_iterables(endpoint,self.endpoints[0]):
            otherIndex = 1
        elif equal_iterables(endpoint,self.endpoints[1]):
            otherIndex= 0
        else:
            raise ValueError("endpoint {} not on this line".format(endpoint))

        # get degree
        deg = angle_between_two_points_clockwise(self.endpoints[otherIndex], endpoint)
        return generate_line_by_length_and_angle(self.endpoints[otherIndex], remLength, deg)

    def trim_to_required_length_from_startpoint(self, requiredLength, endpoint):
        """
        line gets trimmed from .opposing end to start point
        """
        pd = point_distance(self.endpoints[0],self.endpoints[1])
        trimLength = pd - requiredLength
        index = 1 if equal_iterables(endpoint, self.endpoints[0]) else 0

        if trimLength == 0:
            if index:
                return Line([self.endpoints[0],self.endpoints[1]])
            else:
                return Line([self.endpoints[1],self.endpoints[0]])

        l = self.trim_from_end(self.endpoints[index], trimLength)

        return l

    # TODO: test this
    def trim_to_fit_in_area(self,area):
        assert is_valid_area(area), "area invalid"
        if not self.intersects_with_area(area): return None
        np0, np1 = np.copy(self.endpoints[0]), np.copy(self.endpoints[1])

        def check_endpoint_for_bounds_violation(e):
            # case: x out of range
            if e[0] < area[0,0]:
                newY = self.y_given_x(area[0,0])
                e = np.array([area[0,0], newY])
            elif e[0] > area[1,0]:
                newY = self.y_given_x(area[1,0])
                e = np.array([area[1,0], newY])

            # case: y out of range
            if e[1] < area[0,1]:
                newX = self.x_given_y(area[0,1])
                e = np.array([newX, area[0,1]])
            if e[1] > area[1,1]:
                newX = self.x_given_y(area[1,1])
                e = np.array([newX, area[1,1]])
            return e

        np0,np1 = check_endpoint_for_bounds_violation(np0), check_endpoint_for_bounds_violation(np1)
        return Line(np.array([np0,np1]))

    def trim_to_not_fit_in_area(self, area):
        q = self.trim_to_fit_in_area(area)
        if q == None: return [self]

        cs = complement_of_range_in_range(tuple(self.endpoints[:,0]), tuple(q.endpoints[:,0]))
        ls_ = []
        for c in cs:
            p1 = (c[0], self.y_given_x(c[0]))
            p2 = (c[1], self.y_given_x(c[1]))
            p = Line(np.array([p1,p2]))
            ls_.append(p)

        return ls_

    def union_with_bounds(self, area):
        xBound = (min(self.endpoints[0,0], self.endpoints[1,0]), max(self.endpoints[0,0], self.endpoints[1,0]))
        yBound = (min(self.endpoints[0,1], self.endpoints[1,1]), max(self.endpoints[0,1], self.endpoints[1,1]))

        xInt = range_intersection(xBound, (area[0,0], area[1,0]))
        yInt = range_intersection(yBound, (area[0,1], area[1,1]))
        return xInt and yInt

    def point_in_triangle(self, a,b,c):
        '''
        line either
        - exists completely in triangle
        - intersects w/ triangle edge
        '''

        # case: completely in triangle, check center pt.
        if point_in_triangle(a,b,c,self.center):
            return True

        # case: intersection
        l_ = Line([a,b])
        _,stat = intersection_between_two_line_segments(self,l_)
        if stat:
            return True

        l_ = Line([a,c])
        _,stat = intersection_between_two_line_segments(self,l_)
        if stat:
            return True

        l_ = Line([b,c])
        _,stat = intersection_between_two_line_segments(self,l_)
        if stat:
            return True

        return False

    def point_in_circle(self, centerPoint, radius):
        '''
        similar to above, but w/ circle.
        '''

        # case: completely in circle, check center pt.
        if is_point_in_circle(self.center, centerPoint, radius):
            return True

        # case: intersects w/ circle, check both endpoints
        if is_point_in_circle(self.center, self.endpoints[0], radius):
            return True

        if is_point_in_circle(self.center, self.endpoints[1], radius):
            return True

        return False

    def intersects_with_area(self, area):
        if not self.union_with_bounds(area): return False
        NONE,i = self.intersection_point_with_area(area)
        return True if i != -1 else False

    # TODO: untested.
    """
    return:
    - point
    - index::(edge)
    """
    def intersection_point_with_area(self, area):
        assert is_valid_area(area), "invalid area"
        # vertices
        """
        v1  v0
        v2  v3
        """
        corners = area_to_corners(area)

        # make a copy
        for i in range(4):
            l = Line(np.array([corners[i], corners[(i + 1) %4]]))
            point, stat = intersection_between_two_line_segments(l, self)
            if stat:
                return point,i
        return None,-1

    def subtract_line(self, l):
        e1 = np.copy(self.endpoints[0])
        e2 = np.copy(l.endpoints[1])
        return Line(np.array(e1,e2))

    def cross_product_line(self, l):
        assert equal_iterables(l.endpoints[0], self.endpoints[0]), "invalid iterables"

        # get angle b/t lines
        angle1 = angle_between_two_points_clockwise(self.endpoints[0], self.endpoints[1])
        angle2 = angle_between_two_points_clockwise(l.endpoints[0], l.endpoints[1])
        angle = min((angle1 - angle2) % 360, 360 - (angle1 - angle2) % 360)

        d1 = point_distance(self.endpoints[0], self.endpoints[1])
        d2 = point_distance(l.endpoints[0], l.endpoints[1])
        return d1 * d2 * math.sin(math.radians(angle))

    # TODO: unused.
    def move_line_by_y_intercept(self, increment = 0.05):
        p1 = self.slope * self.endpoints[0,0] + self.yint + increment
        p2 = self.slope * self.endpoints[1,0] + self.yint + increment
        return Line([p1, p2])

    def move_line_by_xcoord(self, increment):
        newEndpoints = np.array([[self.endpoints[0,0] + increment, self.endpoints[0,1]], [self.endpoints[1,0] + increment, self.endpoints[1,1]]])
        return Line(newEndpoints)

    def move_line_by_axis_shifts(self,shifts):
        assert is_valid_point(shifts), "invalid shifts"
        np1 = (self.endpoints[0,0] + shifts[0], self.endpoints[0,1] + shifts[1])
        np2 = (self.endpoints[0,0] + shifts[0], self.endpoints[0,1] + shifts[1])
        return Line(np.array([np1,np2]))

    # TODO: test this
    def complement_of_subline_segment(self, l):

        assert self.is_point_in_line_segment(l.endpoints[0])\
            and self.is_point_in_line_segment(l.endpoints[1]), "invalid subline {}".format(l)

        r1 = np.copy(self.endpoints[:,0])
        r2 = np.copy(l.endpoints[:,0])

        complement = complement_of_range_in_range(r1, r2)
        lines = []

        for c in complement:
            p1 = (c[0], self.y_given_x(c[0]))
            p2 = (c[1], self.y_given_x(c[1]))
            p = np.array([p1,p2])
            l2 = Line(p)
            lines.append(l2)
        return lines

    def perpendicular_line(self, radius = 10, source = "mid"):
        """
    default @ midpoint
        """
        if source == "mid":
            source = tuple(self.center)
        else:
            assert is_valid_point(source), "source is not point"
            assert self.is_point_in_line_segment(source), "source is not in line segment"

            y = self.y_given_x(sourceX)
            source = (sourceX, y)

        angle = angle_between_two_points_clockwise(self.endpoints[0],self.endpoints[1])

        neg = (angle - 90.0) % 360.0
        pos = (angle + 90.0) % 360.0

        q = generate_line_by_length_and_angle(source, radius, neg)
        q2 = generate_line_by_length_and_angle(source, radius, pos)
        return Line(np.array([q.endpoints[1], q2.endpoints[1]]))

    def perpendicular_line_in_area_halfspace(self, halfspace):
        assert halfspace in [-1,1], "invalid halfspace {}".format(halfspace)
        assert is_valid_area(area), "invalid area"
        mid = midpoint(self.endpoints[0], self.endpoints[1])

        degDiff = 90.0 * halfspace
        deg = angle_between_two_points_clockwise(self.endpoints[0], self.endpoints[1])

        # form area
        area = two_points_to_area(self.endpoints)

        # get diagonal length total
        diagLength = point_distance(area[0], area[1])
        boundX = area[0,0] if halfSpace < 0 else area[1,0]

        # draw perp line
        q = generate_line_by_length_and_angle(mid, diagLength, deg + degDiff)
        return self.trim_to_fit_in_area(area)

    def gather_variables(self, calculateCenter):
        self.slope = point_slope(self.endpoints[0], self.endpoints[1])
        self.yint = self.endpoints[0][1] - self.endpoints[0][0] * self.slope
        self.center = None
        if calculateCenter:
            self.center = midpoint(self.endpoints[0], self.endpoints[1])

    # TODO: not tested.
    def point_from_source_by_value(self, value, sourceIndex):
        assert value >= 0 and value <= 1, "invalid value"
        assert sourceIndex in [0,1], "invalid source index"
        l = value * point_distance(self.endpoints[0], self.endpoints[1])
        p = self.endpoints[sourceIndex]
        d = angle_between_two_points_clockwise(self.endpoints[0],self.endpoints[1])
        l2 = generate_line_by_length_and_angle(p,l,d)
        return l2.endpoints[1]

    # TODO: check for vertical lines
    def y_given_x(self, x, rangeCheck = True):
        xs = [self.endpoints[0][0], self.endpoints[1][0]]
        if rangeCheck:
            assert not (x < min(xs) or x > max(xs)), "x not in range of line, {}, actual {}".format(x,self.endpoints)
        return round(self.slope * x + self.yint,5) ###?

    # TODO: check for horizontal lines
    def x_given_y(self, y, rangeCheck = True):
        ys = [self.endpoints[0][1], self.endpoints[1][1]]
        if rangeCheck:
            assert not (y < min(ys) or y > max(ys)), "y {} not in range of line {}".format(y,ys)

        if self.slope == 0:
            return self.endpoints[0,0]
        return round((y - self.yint) / float(self.slope),5) ###?

    def is_point_in_line__approximation(self, p):
        '''
        approximation method by point distance
        '''

        p2 = np.round((p[0], self.y_given_x(p[0], False)),5)

        # get point distance
        if point_distance(p, p2) <= 10 ** -2:
            return True

        return False

    def is_point_in_line_segment(self, p):

        # case: horizontal line
        if self.is_horizontal():
            return True if (p[1] - self.endpoints[0,1]) <= 10 ** -5 and\
                self.value_in_coordinate_range(p[0],0) else False

        # case: vertical line
        elif self.is_vertical():
            return True if (p[0] - self.endpoints[0,0]) <= 10 ** -5 and\
                self.value_in_coordinate_range(p[1],1) else False

        # case: other
        p2 = np.round((p[0], self.y_given_x(p[0], False)),5)

        stat = self.is_point_in_line__approximation(p)
        stat2 = self.value_in_coordinate_range(p2[0], 0)
        stat3 = self.value_in_coordinate_range(p2[1], 1)

        return stat and stat2 and stat3

    def value_in_coordinate_range(self, v, coord):
        q = self.endpoints[:,coord]
        return True if v >= min(q) and v <= max(q) else False

    def form_point_sequence(self, output = "exact", pointHop = DEFAULT_TRAVELLING_HOP):## 0.1):
        assert output in ["integer", "exact"]
        assert pointHop > 0, "invalid point hop"

        def get_point(v):
            if a == 0:
                return [int(v), int(self.y_given_x(start))] if output == "integer" else [start, self.y_given_x(start)]
            else:
                return [int(self.x_given_y(start)), int(v)] if output == "integer" else [self.x_given_y(start), start]

        # determine axis of hop
        a = self.axis_with_greater_change()

        if self.endpoints[0][a] < self.endpoints[1][a]:
            start, end = self.endpoints[0][a], self.endpoints[1][a]
        else:
            start, end = self.endpoints[1][a], self.endpoints[0][a]

        self.data = []

        while start < end:
            self.data.append(get_point(start))
            start += pointHop

        self.data = np.array(self.data)

    def coordinate_directions_given_endpoints(self):
        """
        determines the (x,y) directions of line given its endpoints and slope.
        """

        xDir,yDir = None,None
        xDir = 1 if self.endpoints[0][0] <= self.endpoints[1][0] else -1
        yDir = 1 if self.slope >= 0 else -1
        return xDir, yDir

    # TODO: unused, delete.
    def center_xcoord_at_zero(self):
        length = abs(self.endpoints[0,0] - self.endpoints[1,0])

        d = min(self.endpoints[:,0])
        sd = d + length / 2.0
        newEndpoints = np.array([[self.endpoints[0,0] - sd, self.endpoints[0,1]], [self.endpoints[1,0] - sd, self.endpoints[1,1]] ])
        return Line(newEndpoints)


def intersection_between_two_lines(l1,l2):

    # case: identical lines
    if l1 == l2:
        return l1.endpoints[0], True

    # case: one of the lines is vertical,
    v1,v2 = l1.is_vertical(), l2.is_vertical()

    if v1 or v2:
        if v1 and v2:
            return None,False

        xTarg = l1.endpoints[0,0] if v1 else l2.endpoints[0,0]
        p = (xTarg, l1.y_given_x(xTarg, False)) if v2 else (xTarg, l2.y_given_x(xTarg, False))
        return p, True

    # solve equation: m_1 * x + b = m_2 * x + b_2
    '''
    (m_1 - m_2) * x = b_2 - b

    x = (b_2 - b) / (m_1 - m_2)
    '''
    if abs(l1.slope - l2.slope) < 10 ** -5 and abs(l1.yint - l2.yint) > 10 ** -5:
        return None, False

    x = float(l2.yint - l1.yint) / float(l1.slope - l2.slope)
    y = l1.y_given_x(x, False)
    return (x,y), True

def intersection_between_two_line_segments(l1,l2):
    q, stat = intersection_between_two_lines(l1,l2)

    if not stat: return q,stat
    if not l1.is_point_in_line_segment(q): return None, False
    if not l2.is_point_in_line_segment(q): return None, False
    return q,stat

def point_in_triangle(a,b,c,p):
    if equal_iterables(a, p) or equal_iterables(b,p) or equal_iterables(c,p):
        return True

    lab = Line(np.array([a,b]))
    lac = Line(np.array([a,c]))
    lap = Line(np.array([a,p]))
    abAng = angle_between_two_lines(lab,lac)
    apAng = angle_between_two_lines(lab, lap)

    if apAng > abAng:
        return False

    # get point of intersection b/t
    lbc = Line(np.array([b,c]))
    point,stat = intersection_between_two_lines(lbc, lap)

    # check distance
    apDist = point_distance(a, p)
    aBoundDist = point_distance(a, point)

    if apDist <= aBoundDist: return True
    return False

"""
"""
def angle_between_two_lines(l1, l2):

    assert equal_iterables(l1.endpoints[0], l2.endpoints[0]), "invalid iterables"

    # get angle b/t lines
    angle1 = angle_between_two_points_clockwise(l1.endpoints[0], l1.endpoints[1])
    angle2 = angle_between_two_points_clockwise(l2.endpoints[0], l2.endpoints[1])
    angle = min((angle1 - angle2) % 360, 360 - (angle1 - angle2) % 360)
    return angle

# TODO: un-used
def tangent_of_line(line, point):
    assert is_valid_point(point)

    # choose another x
    xp = point[0] + 10
    q = y - (1 / line.slope * (xp - point[0]))
    point2 = (xp, q)
    return Line(point, point2)



from .rssi import *

DEFAULT_SINGLE_WRITE_SIZE = 2000

'''
fu
'''
def one_random_noise_(parentBounds,bounds, noiseRange):
    if not is_proper_bounds_vector(bounds):
        pd = point_difference_of_improper_bounds(parentBounds,bounds)
        s1 = np.zeros((parentBounds.shape[0],))
        bx = np.array([s1,pd]).T
    else:
        bx = bounds
    return one_random_noise(bx,noiseRange)

class NSDataInstructions:
    '''
    Data structure that uses `ResplattingSearchSpaceIterator` to generate data.

    If arg<noiseRange> != None, then add noise to each point value.
    Generates data into file according to instructions given by arguments.

    :param bInf: (bounds, startPoint, columnOrder, SSIHop, additionalUpdateArgs)
    :param rm: (mode := `relevance zoom` | `prg` | sequence::(relevant bounds), RCH)
    :param filePath: str
    :param noiseRange: n x 2
    :type noiseRange: np.ndarray
    :param writeOutMode: "literal" to write out every point iterated,
                        "relevant" to write out only points deemed relevant by RCH,
                        RCH to write out RCH(p) for every point p. 
    '''

    def __init__(self, bInf, rm,filePath,modeia,noiseRange = None,writeOutMode = "literal"):
        self.bInf = bInf
        self.rm = rm
        self.filePath = filePath
        self.fp = None
        self.load_filepath(modeia)
        self.nr = noiseRange
        assert writeOutMode in {'literal','relevant'} or type(writeOutMode) is RChainHead
        self.wom = writeOutMode
        self.rchPrev = None # used for writeOutMode == 'relevant'

        self.c = 0
        self.terminated = False
        self.bs = []
        return

    def load_filepath(self,modeia):

        # folder
        if "/" in self.filePath:
            # check exists
            #
            s = self.filePath[::-1]
            q = s.find('/')
            s = s[q + 1:][::-1]
            print("FP ",s)

            # make dir
            if not os.path.isdir(s):
                # make directory
                os.mkdir(s)
                modeia = 'w'

        self.fp = open(self.filePath,modeia)
        return

    def make_rssi(self):
        # mock a delaani
        if type(self.rm[0]) != str:
            delaani = ("relevance zoom",self.rm[1])
        else:
            delaani = self.rm

        #bounds,star
        self.rssi = ResplattingSearchSpaceIterator(self.bInf[0], self.bInf[1],\
                self.bInf[2],self.bInf[3],delaani,additionalUpdateArgs = self.bInf[4])
        return

    def next_batch(self):
        # load next bound in self.rm[0]
        if type(self.rm[0]) != str:
            delaani = ("relevance zoom",self.rm[1])
        else:
            delaani = self.rm

        # update the instructor if not `relevance zoom` or `prg`
        if type(self.rm[0]) != str and self.c > 1:
            if len(self.rm[0]) == 0:
                self.terminated = True
                print("DONYA")
                return None

            q = self.rm[0][0]
            x = self.rm[0][1:]
            self.rm = (x,self.rm[1])
            #print("RMMM")
            #print(self.rm)

            # start point is left
            DEFAULT_START_POINT = np.copy(q[:,0])
            self.rssi.update_resplatting_instructor((q,DEFAULT_START_POINT))

        if self.rssi.terminated:
            return None

        if self.wom == "relevant":
            self.rchPrev = deepcopy(self.rm[1].apply)

        # fetch the bound
        if delaani[0] == "relevance zoom":
            dsws = DEFAULT_SINGLE_WRITE_SIZE
            q = ResplattingSearchSpaceIterator.iterate_one_batch(self.rssi,dsws)
        else: # prg
            q = []
            qc = 0

            while qc < DEFAULT_SINGLE_WRITE_SIZE:
                nx = next(self.rssi)
                if type(nx) == None:
                    break
                q.append(nx)
                qc += 1

        # filter out by relevance function
        if self.wom == "relevant":
            q = self.relevance_filter(q)
            ##print("LEN ", len(q))
        elif type(self.wom) is RChainHead:
            q = self.wom_rch_map(q) 

        return q

    def relevance_filter(self,q):
        x = []
        for q_ in q:
            if self.rchPrev(q_):
                x.append(q_)
        self.rchPrev = None
        return x

    def wom_rch_map(self,q):
        self.update_wom_rch() 
        return [self.wom.apply(q_) for q_ in q]

    def update_wom_rch(self):
        q = self.rssi.update_vars_for_rch()
        self.wom.load_update_vars(q)
        self.wom.update_rch()


    def add_noise_to_point(self,p):

        h = one_random_noise_(self.rssi.bounds,\
                self.rssi.ssi.de_bounds(),\
                self.nr)

        if type(self.rssi.ssi) is SkewedSearchSpaceIterator:
            p_ = self.rssi.ssi.inverse_round_value(p)
            p_ = vector_hop_in_bounds(p_,h,self.rssi.ssi.iBounds)
            return self.rssi.ssi.round_value(p_)
        else:
            return vector_hop_in_bounds(p,h,self.rssi.ssi.de_bounds())

    # TODO: untested
    def add_noise_to_batch(self,b):
        if type(self.nr) == type(None):
            return b

        for p in b:
            yield self.add_noise_to_point(p)

    def next_batch_(self):
        self.c += 1
        if type(self.fp) == type(None):
            return

        q = self.next_batch()
        if type(q) != type(None):
            if type(self.rm[0]) == str and self.rm[0] == 'relevance zoom':
                b = np.copy(self.rssi.ssi.bounds)
            else:
                b = np.copy(self.rssi.bounds)

            # check for adding noise
            if type(self.nr) != type(None) and type(self.wom) != RChainHead:
                q2 = self.add_noise_to_batch(q) 
                q = q2

            q = [vector_to_string(q_,cr) + "\n" for q_ in q]
            self.fp.writelines(q)
            self.fp.flush()

            # summarize
            l = len(q)
            self.bs.append([b,l])
        else:
            self.close()

    def batch_summary(self):
        for (i,bs) in enumerate(self.bs):
            print("batch #",i)
            print("- bound")
            print(bs[0])
            print("- size ",bs[1])
            print()

    def close(self):
        self.fp.close()
        self.fp = None



'''
methods on integers such as multiple-finding
'''

def all_multiples(i):
    l = abs(int(i / 2))
    m = set()
    for j in range(1,l + 1):
        if j in m: break
        if not i % j:
            m |= set([j,int(i/j)])
    return m

def all_multiple_pairs(i):
    if i == 1: return [(1,1)]
    if i == -1: return [(-1,1),(1,-1)]
    
    l = abs(int(i / 2))
    m = []
    s = set()
    for j in range(1,l + 1):
        if not i % j:
            if j in s: break
            if int(i/j) in s: break
            m.append((j,int(i/j)))
            m.append((int(i/j),j))
            s |= set([j,int(i/j)])

            # case: negative
            if i < 0:
                s |= set([-j,-int(i/j)])
                m.append((-j,-int(i/j)))
                m.append((-int(i/j),-j))            
    return m

def gcd(i1,i2):
    i1 = to_natural_number(i1)
    i2 = to_natural_number(i2)
    q = max([int(min([i1,i2]) / 2),2])
    m = int(min([i1,i2]))
    c = 1
    for i in range(1,q):
        if not i1 % i and not i2 % i\
            and i > c:
            c = i
        
        if not m % i:
            if not i1 % (m / i) and not\
                i2 % (m / i) and (m / i) > c:
                c = int(m / i)
    return c

def to_natural_number(i):
    if type(i) == complex:
        return to_natural_number(int((i / 1j).real))
    if i < 0:
        return i * -1
    return i

def fractional_add(f1,f2):
    if f1[1] == f2[1]:
        s = (f1[0] + f2[0],f2[1])
        return reduce_fraction(s) 
    f1_ = (f1[0] * f2[1],f1[1] * f2[1])
    f2_ = (f2[0] * f1[1],f2[1] * f1[1])
    return fractional_add(f1_,f2_)

def fractional_mul(f1,f2):
    f3 = (f1[0] * f2[0],f1[1] * f2[1])
    return reduce_fraction(f3)

def reduce_fraction(f):
    d = gcd(f[0],f[1])
    return (int(f[0] / d), int(f[1] / d))

def is_rational(f):
    assert f >= -1. and f <= 1., "invalid float"
    return "TODO"




class ViolationHandler1:
    """
    Decision structure for :class:`BallComp` algorithm.

    ViolationHandler1 is initialized with a `max balls` and `max radius`
    argument. If BallComp uses this handler, then BallComp will terminate once
    either of the max conditions of this handler is reached.
    """

    def __init__(self,maxBalls,maxRadius):
        self.maxBalls = maxBalls
        self.maxRadius = maxRadius
        return

    def check_violation_1(self,ballRadius, maxBallCompRadius):
        """
        :return: terminate,delta value
        :rtype: bool,float
        """

        if ballRadius >= self.maxRadius - 0.05:
            return (True,None)

        if ballRadius <= maxBallCompRadius:
            return (False,None)

        q = self.maxRadius - maxBallCompRadius
        d = ballRadius - maxBallCompRadius
        return (False,d + (q / 2.0))

    """
    return:
    - (bool::terminate,float::(delta value))
    """
    def check_violation_2(self, numberOfBalls, ballCompMaxBalls):
        print("CV {} {} {}".format(numberOfBalls,ballCompMaxBalls, self.maxBalls))

        # case: terminate
        if numberOfBalls > self.maxBalls:
            return (True,None)

        if numberOfBalls <= ballCompMaxBalls:
            return (False,None)

        d = numberOfBalls - ballCompMaxBalls
        print("CV2 {} {}".format(numberOfBalls,d))

        # case: spare balls from .`maxBalls`, add 2 extra
        if numberOfBalls + 2 <= self.maxBalls:
            return (False,d + 2)

        # case: spare balls from .`maxBalls`, add 1 extra
        if numberOfBalls + 1 <= self.maxBalls:
            return (False,d + 1)

        # case: no more spare balls
        return (False,d)



from .graph_basics_test_cases import *
from morebs2 import graph_basics 
import unittest


def nx_mapcheck_with_functions(D,nx_map,is_np:bool): 
    """
    nx map is either neighbor-child or neighbor-parent 
    map 
    """

    nf = graph_basics.doubly_connected
    f = graph_basics.parents_of if is_np else graph_basics.children_of

    for k,v in nx_map.items(): 
        assert v[0] == len(nf(D,k))
        assert v[1] == len(f(D,k))
    return 

'''
python -m morebs2.tests.graph_basics_test  
'''
class TestGraphBasicsMethods(unittest.TestCase):

    def test__GraphBasics__connections_measures_case1(self):
        D = graph_case_1()
        C = graph_basics.connected_to(D,0)
        assert C == set([1,2,3])
        P = graph_basics.parents_of(D,0)
        assert P == set([2,3])
        C2 = graph_basics.children_of(D,0)
        assert C2 == set([])
        D2 = graph_basics.doubly_connected(D,0)
        assert D2 == set([1])
        return

    def test__GraphBasics__connections_measures_case2(self):
        D = graph_case_8() 
        C = graph_basics.connected_to(D,7)
        assert C == set([0,3,8]) 
        P = graph_basics.parents_of(D,7) 
        assert P == set() 
        C2 = graph_basics.children_of(D,7) 
        assert C2 == set([0]) 
        D2 = graph_basics.doubly_connected(D,7)
        assert D2 == set([3,8])

    def test__GraphBasics__connections_measures_case3(self):
        D = graph_case_9() 
        prt = graph_basics.directed_edge_partition(D,1,D[1]) 
        assert prt == [set([0]),set([5,6])]

        D = graph_case_4()
        P = graph_basics.directed_edge_partition(D,0,[1,2])
        assert P == [set(), {1, 2}]

    def test__GraphBasics__connections_measures_case4(self):
        D = graph_case_9() 
        assert graph_basics.edge_count(D) == 11 

        D = graph_case_2() 
        assert graph_basics.edge_count(D) == 4 

        D = graph_case_3() 
        assert graph_basics.edge_count(D) == 8 

        D = graph_case_1() 
        assert graph_basics.edge_count(D) == 5 

    def test__GraphBasics__connections_measures_case5(self):

        D = graph_case_10()
        assert not graph_basics.is_directed_graph(D)

        D = graph_case_4()
        assert graph_basics.is_directed_graph(D)

        D = graph_case_2()
        assert graph_basics.is_directed_graph(D)

        D = graph_case_3()
        assert not graph_basics.is_directed_graph(D)

    def test__GraphBasics__connections_measures_case6(self):
        D = graph_case_12() 

        nc = graph_basics.nc_degree_map(D)
        npx = graph_basics.np_degree_map(D) 
        nx_mapcheck_with_functions(D,nc,False)
        nx_mapcheck_with_functions(D,npx,True) 

    def test__GraphBasics__connections_measures_case7(self):
        D = graph_case_1() 

        nc = graph_basics.nc_degree_map(D)
        npx = graph_basics.np_degree_map(D) 
        nx_mapcheck_with_functions(D,nc,False)
        nx_mapcheck_with_functions(D,npx,True) 

        assert nc == defaultdict(list, \
            {0: [1, 0], 1: [1, 1], 2: [0, 1], 3: [0, 1]})
        assert npx == defaultdict(list, \
            {0: [1, 2], 1: [1, 0], 2: [0, 1], 3: [0, 0]})

    def test__GraphBasics__connections_measures_case8(self):
        D = graph_case_2() 

        nc = graph_basics.nc_degree_map(D)
        npx = graph_basics.np_degree_map(D) 
        nx_mapcheck_with_functions(D,nc,False)
        nx_mapcheck_with_functions(D,npx,True) 

        assert nc == defaultdict(list, \
            {0: [0, 0], 1: [0, 1], 2: [0, 1], 3: [0, 1], 4: [0, 1]})
        assert npx == defaultdict(list, \
            {0: [0, 4], 1: [0, 0], 2: [0, 0], 3: [0, 0], 4: [0, 0]})



class TestGraphComponentDecompositionClass(unittest.TestCase):

    def test__GraphComponentDecomposition__init_decomp(self):
        D = graph_case_2() 
        gd = graph_basics.GraphComponentDecomposition(D)
        gd.init_decomp(0) 
        assert gd.components == [[{0}]]

        D = graph_case_3()
        gd = graph_basics.GraphComponentDecomposition(D)
        gd.is_directed = True 
        gd.init_decomp(0) 
        assert gd.components == [[{0,1,2,3,4}]]
        assert gd.key_cache == {0}
        assert gd.key_queue == [1, 2, 3, 4]
        gd.next_key()
        assert gd.key_queue == [2, 3, 4]
        assert gd.components == [[{0,1,2,3,4}]]

    def test__GraphComponentDecomposition__next_key_case1(self):
        D = graph_case_4()
        gd = graph_basics.GraphComponentDecomposition(D)
        gd.next_key()
        assert gd.components == [[{0}, {1}], [{0}, {2}]]
        gd.next_key()
        assert gd.components == [[{0}, {2}], [{0}, {1}, {3}], [{0}, {1}, {4}]]
        while not gd.finstat:
            gd.next_key()
        assert gd.components == \
            [[{0}, {1}, {4}], [{0}, {2}, {5}], [{0}, {2}, {6}], [{0}, {1}, {3, 7}]]

    def test__GraphComponentDecomposition__next_key_case2(self):
        D = graph_case_5()

        P = graph_basics.directed_edge_partition(D,0,[1,2])
        assert P == [set(), {1, 2}]

        gd = graph_basics.GraphComponentDecomposition(D)
        gd.next_key()
        assert gd.components == [[{0}, {1}], [{0}, {2}]]
        gd.next_key()
        assert gd.components == [[{0}, {2}], [{0}, {1}, {3}], [{0}, {1}, {4}]]

    def test__GraphComponentDecomposition__next_key_case3(self):
        D = graph_case_6()
        gd = graph_basics.GraphComponentDecomposition(D)
        gd.is_directed = True 
        while not gd.finstat:
            gd.next_key()

        assert gd.components == [[{0, 1, 2, 3, 4, 5, 6}]]

    def test__GraphComponentDecomposition__next_key_case4(self):
        D = graph_case_7()
        gd = graph_basics.GraphComponentDecomposition(D)
        while not gd.finstat:
            gd.next_key()

        assert graph_basics.edge_count(gd.d_) == 0 
        assert gd.components == \
            [[{0}, {1}, {4}], [{0}, {2}, {5}], \
            [{0}, {2}, {6}], \
            [{0}, {1}, {8, 9, 3, 7}, {10}]]
        assert gd.cyclic_keys == defaultdict(set,{7: {0}})
        assert gd.cyclic_component_indices() == [3]

        drm = defaultdict(int, {0: 0, 1: 2, 4: 2, \
            2: 2, 5: 2, 6: 2, 8: 2, 9: 2, 3: 2, 7: 2, 10: 3})
        assert gd.depth_rank_map() == drm 
        return

    def test__GraphComponentDecomposition__next_key_case5(self):
        D = graph_case_8()
        gd = graph_basics.GraphComponentDecomposition(D)
        while not gd.finstat:
            gd.next_key()

        assert graph_basics.edge_count(gd.d_) == 0 

        assert gd.components == \
            [[{0}, {1}, {4}], [{0}, {2}, {5}], \
            [{0}, {2}, {6}], \
            [{0}, {1}, {3, 7, 8, 9, 10}]]


        gd = graph_basics.GraphComponentDecomposition(D) 
        gd.init_decomp(9) 
        while not gd.finstat:
            gd.next_key()

        assert gd.components == [[{3, 7, 8, 9, 10}, {0}, {1}, {4}], \
            [{3, 7, 8, 9, 10}, {0}, {2}, {5}], \
            [{3, 7, 8, 9, 10}, {0}, {2}, {6}]]

        return 

    def test__GraphComponentDecomposition__next_key_case6(self):

        D = graph_case_9()
        gd = graph_basics.GraphComponentDecomposition(D)
        while not gd.finstat:
            gd.next_key()
        assert gd.cyclic_keys == defaultdict(set, {5: {0}})
        assert gd.components == [[{0, 1, 2, 3, 4}, {5}], [{0, 1, 2, 3, 4}, {6}]]

    def test__GraphComponentDecomposition__next_key_case7(self):
        D = graph_case_10()
        gd = graph_basics.GraphComponentDecomposition(D)

        while not gd.finstat: 
            gd.next_key()

        assert gd.components == [{0, 1, 2, 3, 4}, {7, 8, 9}, {10, 111}]

    def test__GraphComponentDecomposition__next_key_case8(self):
        D = graph_case_11() 
        gd = graph_basics.GraphComponentDecomposition(D) 
        gd.decompose()
        assert gd.components == [[{0, 1, 2, 3, 4}, {5}], \
            [{0, 1, 2, 3, 4}, {6}], [{7}, {12}, {41}], [{7}, {20}, {64}], \
            [{7}, {20}, {15}], [{7}, {12}, {38}, {11}, {31}]], "got {}".format(gd.components)

    def test__GraphComponentDecomposition__next_key_case9(self):

        D = graph_case_2()
        gd = graph_basics.GraphComponentDecomposition(D) 
        gd.decompose() 

        dr = gd.depth_rank_map() 
        assert dr == defaultdict(int, \
            {0: 4, 1: 0, 2: 0, 3: 0, 4: 0})

        assert gd.components == \
            [[{0}], [{1}, {0}], [{2}, {0}], [{3}, {0}], [{4}, {0}]]

    def test__GraphComponentDecomposition__next_key_case10(self):

        D = graph_case_3()
        gd = graph_basics.GraphComponentDecomposition(D) 
        gd.decompose() 

        ci = gd.cyclic_component_indices()
        dr = gd.depth_rank_map() 

        assert gd.components == [{0, 1, 2, 3, 4}]
        assert type(dr) == type(None) 

    def test__GraphComponentDecomposition__next_key_case11(self):
        D = graph_case_13() 
        gd = graph_basics.GraphComponentDecomposition(D)
        gd.decompose()
        assert gd.components == [[{0, 7}, {9, 6}, {8}, {10}]]
        assert gd.cyclic_keys == defaultdict(set, {8: {0}})

    def test__GraphComponentDecomposition__dcomponent_cyclic_indices(self): 

        D = graph_case_14() 
        gd = graph_basics.GraphComponentDecomposition(D)
        gd.decompose()
        assert gd.dcomponent_cyclic_indices(0) == [(0, 2), (3, 5)]
        q = gd.depth_rank_map__component(0)
        assert q == defaultdict(int, {0: 0, 7: 0, \
            9: 0, 6: 0, 8: 0, 10: 1, 12: 1, \
            11: 2, 13: 2})

        D[13].clear() 
        gd = graph_basics.GraphComponentDecomposition(D)
        gd.decompose()
        assert gd.dcomponent_cyclic_indices(0) == [(0, 2)]
        q = gd.depth_rank_map__component(0)
        assert q == defaultdict(int, {0: 0, \
            7: 0, 9: 0, 6: 0, 8: 0, \
            10: 1, 12: 1, 11: 2, 13: 3})

    def test__GraphComponentDecomposition__depth_rank_map__component(self): 
        D = graph_case_2() 
        gd = graph_basics.GraphComponentDecomposition(D)
        gd.decompose()
        q = gd.depth_rank_map__component(0)
        assert q == defaultdict(int,{0:0})
        q2 = gd.depth_rank_map__component(1)
        assert q2 == defaultdict(int, {1: 0, 0: 1})
        return 


if __name__ == '__main__':
    unittest.main()



from .line import *
import random
################ START: line generators

def generate_random_line(startPoint, length):
    """
    generates a random line of `length` from .`startPoint` by a random angle in
    [0,360].
    """
    # random angle
    angle = random.uniform(0, 360)
    endPoint = hypotenuse_from_point(startPoint, length, angle)
    return Line((startPoint, endPoint))

def generate_random_line_at_center(centerPoint, length):
    """
    generates a random line of `length` at center `centerPoint`.

    :param centerPoint: 2-tupules de duos
    :type centerPoint: iterable
    :param length: lengua
    :type length: float
    """
    # random angle
    angle = random.uniform(0, 360)
    altAngle = (angle + 180.0) % 360.0

    e1 = hypotenuse_from_point(centerPoint, length / 2.0, angle)
    e2 = hypotenuse_from_point(centerPoint, length / 2.0, altAngle)
    return Line((e1, e2))

################ END: line generators

def generate_line_by_length_and_angle(startPoint, length, angle):
    endPoint = hypotenuse_from_point(startPoint, length, angle)
    return Line((startPoint, endPoint))

def hypotenuse_from_point(point, length, angle):
    """
    Outputs an endpoint given `point`; `endpoint` is distance `length` from .`point` and at `angle`
    """

    # get the x-delta and y-delta
    q = math.sin(math.radians(angle))
    opp = q * length

    q = math.cos(math.radians(angle))
    adj = q * length

    return [point[0] + adj, point[1] + opp]

    ##################################################################

# TODO: necessary??
"""
"""
def closest_right_angle_to_angle(angle):
    angle = angle % 360
    right = [0, 90, 180, 270, 360]
    diff = [abs(angle - r) for r in right]
    index = np.argmin(diff)
    return right[index]

"""
"""
def quadrant_of_angle(angle):
    """
    determinini es angle y quadrillas
    """

    assert not (angle < 0 or angle > 360), "invalid angle"

    if (angle >= 0 and angle <= 90) or angle == 360:
        return 0

    if (angle > 90 and angle <= 180):
        return 1

    if (angle > 180 and angle <= 270):
        return 2

    return 3

############################### END: line generator

################ START: area identification

def random_point_in_area(area):
    assert is_valid_area(area), "area is invalid"
    x = random.uniform(area[0,0], area[1,0])
    y = random.uniform(area[0,1], area[1,1])
    return (x,y)

def random_point_in_circle(center, radiusRange):
    # random x-delta    
    xDelta = random.uniform(radiusRange[0],radiusRange[1])

    # select a radius range in [xDelta,radiusRange[1]]
    rr = random.uniform(xDelta,radiusRange[1])

    # random y-delta
    yDelta = math.sqrt(rr**2 - xDelta**2)

    s1 = 1 if random.random() >= 0.5 else -1 
    s2 = 1 if random.random() >= 0.5 else -1 
    return (center[0] + xDelta * s1, center[1] + yDelta * s2)

def random_point_near_2d_point_pair(pair,rdistance):
    '''
    generates a random point near a 2d point pair by first
    constructing a line b/t the the points in the pair,
    choosing a random point `r` on the line, and then a random point
    of distance `d` in range `rdistance` from `r`.  
    '''
    assert len(pair) == 2, "invalid pair"
    l = Line([pair[0],pair[1]])

    s = 0
    v = random.random()
    p = l.point_from_source_by_value(v,s)
    return random_point_in_circle(p,rdistance)

def generate_random_xyl_points_at_center(c,drnp,l):
    '''
    for each element in `drnp`, generates 
            by std. python.

    c := center point, (x,y,label)
    drnp := list(<(distance min to c,distance max to c, number of points)>)
    l := int,label
    '''

    def process_drn(drn):
        rps = []
        for i in range(drn[2]):
            p = list(random_point_in_circle(c[:2],drn[:2])) + [l]
            rps.append(p)
        return rps

    ps = []
    for drn in drnp:
        ps.extend(process_drn(drn))
    return np.array(ps)

"""
"""
def random_game_table_matrix(xMoveSize, yMoveSize, rangeX, rangeY):#, rule):
    assert len(rangeX) == 2 and len(rangeY) == 2, "arg. range is wrong"
    q = np.empty((xMoveSize, yMoveSize, 2))

    for i in range(xMoveSize):
        for j in range(yMoveSize):
            q[i,j] = (random.randrange(rangeX[0], rangeX[1]), random.randrange(rangeY[0], rangeY[1]))
    return q
